#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import fixtures
from unittest import mock

from oslo_utils.fixture import uuidsentinel as uuids

from cyborg.common import exception
from cyborg.conductor import manager
from cyborg.tests import base
from cyborg.tests.unit import fake_driver_device


class ConductorManagerTest(base.TestCase):
    def setUp(self):
        super(ConductorManagerTest, self).setUp()
        self.placement_mock = self.useFixture(fixtures.MockPatch(
            'cyborg.common.placement_client.PlacementClient')
        ).mock.return_value
        self.cm = manager.ConductorManager(
            mock.sentinel.topic, mock.sentinel.host)
        self.fake_driver_devices = (fake_driver_device.
                                    get_fake_driver_devices_objs())
        self.fake_driver_depolyables = (fake_driver_device.
                                        get_fake_driver_deployable_objs())

    def test__gen_resource_inventory(self):
        expected = {
            'CUSTOM_FOO': {
                'total': 42,
                'max_unit': 42,
            },
        }
        actual = manager._gen_resource_inventory('CUSTOM_FOO', 42)
        self.assertEqual(expected, actual)

    @mock.patch('cyborg.conductor.manager.ConductorManager._get_sub_provider')
    def test_provider_report(self, mock_get_sub):
        rc = 'CUSTOM_ACCELERATOR'
        traits = ["CUSTOM_FPGA_INTEL",
                  "CUSTOM_FPGA_INTEL_ARRIA10",
                  "CUSTOM_FPGA_INTEL_REGION_UUID",
                  "CUSTOM_FPGA_FUNCTION_ID_INTEL_UUID",
                  "CUSTOM_PROGRAMMABLE",
                  "CUSTOM_FPGA_NETWORK"]
        total = 42
        expected_inv = {
            rc: {
                'total': total,
                'max_unit': total,
            },
        }
        self.placement_mock.ensure_resource_classes.return_value = None
        actual = self.cm.provider_report(
            mock.sentinel.context, mock.sentinel.name, rc, traits, total,
            mock.sentinel.parent)

        self.placement_mock.ensure_resource_classes.assert_called_once_with(
            mock.sentinel.context, [rc])
        mock_get_sub.assert_called_once_with(
            mock.sentinel.context, mock.sentinel.parent, mock.sentinel.name)
        sub_pr_uuid = mock_get_sub.return_value
        self.placement_mock.update_inventory.assert_called_once_with(
            sub_pr_uuid, expected_inv)
        self.placement_mock.add_traits_to_rp.assert_called_once_with(
            sub_pr_uuid, traits)
        self.assertEqual(sub_pr_uuid, actual)

    def test_get_root_provider(self):
        self.placement_mock.get.return_value.json.return_value = {
            'resource_providers': [{'uuid': mock.sentinel.uuid}],
        }
        uuid = self.cm._get_root_provider(mock.sentinel.context, 'foo')
        self.assertEqual(mock.sentinel.uuid, uuid)

    def test_get_root_provider_not_found(self):
        self.placement_mock.get.return_value.json.return_value = {
            'resource_providers': [],
        }
        self.assertRaises(exception.PlacementResourceProviderNotFound,
                          self.cm._get_root_provider,
                          mock.sentinel.context, 'foo')

    def test_get_root_provider_unavailable(self):
        self.placement_mock.get.side_effect = exception.PlacementServerError(
            "Placement Server has some error at this time.")
        self.assertRaises(exception.PlacementServerError,
                          self.cm._get_root_provider,
                          mock.sentinel.context, 'foo')

    @mock.patch('cyborg.conductor.manager.ConductorManager.'
                '_delete_provider_and_sub_providers')
    @mock.patch('cyborg.conductor.manager.ConductorManager.'
                'get_placement_needed_info_and_report')
    @mock.patch('cyborg.objects.driver_objects.driver_device.'
                'DriverDevice.destroy')
    @mock.patch('cyborg.objects.driver_objects.driver_device.'
                'DriverDevice.create')
    def test_drv_device_make_diff(self, mock_create_driver_device,
                                  mock_destroy_driver_device,
                                  mock_placement_report,
                                  mock_placement_delete):
        old_driver_attr_list = []
        new_driver_attr_list = self.fake_driver_devices[:1]
        self.placement_mock.get.return_value.json.return_value = {
            'resource_providers': [{'uuid': mock.sentinel.uuid}],
        }

        mock_placement_report.side_effect = (
            exception.ResourceProviderCreationFailed(
                name=uuids.compute_node))

        self.cm.drv_device_make_diff(
            mock.sentinel.context, 'foo',
            old_driver_attr_list, new_driver_attr_list)

        mock_destroy_driver_device.assert_called_once()
        mock_placement_delete.assert_called_once()

    @mock.patch('cyborg.conductor.manager.ConductorManager.'
                '_delete_provider_and_sub_providers')
    @mock.patch('cyborg.conductor.manager.ConductorManager.'
                'get_placement_needed_info_and_report')
    @mock.patch('cyborg.objects.driver_objects.driver_deployable.'
                'DriverDeployable.destroy')
    @mock.patch('cyborg.objects.driver_objects.driver_deployable.'
                'DriverDeployable.create')
    def test_drv_deployable_make_diff(self, mock_create_driver_deployable,
                                      mock_destroy_driver_deployable,
                                      mock_placement_report,
                                      mock_placement_delete):
        old_driver_dep_list = []
        new_driver_dep_list = self.fake_driver_depolyables[:1]
        self.placement_mock.get.return_value.json.return_value = {
            'resource_providers': [{'uuid': mock.sentinel.uuid}],
        }

        mock_placement_report.side_effect = (
            exception.ResourceProviderCreationFailed(
                name=uuids.compute_node))

        self.cm.drv_deployable_make_diff(
            mock.sentinel.context, '1', '2',
            old_driver_dep_list, new_driver_dep_list, 'foo')

        mock_destroy_driver_deployable.assert_called_once()
        mock_placement_delete.assert_called_once()



# Copyright 2011 OpenStack Foundation
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1.access import service_catalog as ksa_service_catalog
from keystoneauth1 import plugin
from oslo_context import context
from oslo_db.sqlalchemy import enginefacade
from oslo_utils import timeutils

from cyborg.common import exception
from cyborg.common import utils


class _ContextAuthPlugin(plugin.BaseAuthPlugin):
    """A keystoneauth auth plugin that uses the values from the Context.

    Ideally we would use the plugin provided by auth_token middleware however
    this plugin isn't serialized yet so we construct one from the serialized
    auth data.
    """

    def __init__(self, auth_token, sc):
        super(_ContextAuthPlugin, self).__init__()

        self.auth_token = auth_token
        self.service_catalog = ksa_service_catalog.ServiceCatalogV2(sc)

    def get_token(self, *args, **kwargs):
        return self.auth_token

    def get_endpoint(self, session, service_type=None, interface=None,
                     region_name=None, service_name=None, **kwargs):
        return self.service_catalog.url_for(service_type=service_type,
                                            service_name=service_name,
                                            interface=interface,
                                            region_name=region_name)


@enginefacade.transaction_context_provider
class RequestContext(context.RequestContext):
    """Security context and request information.

    Represents the user taking a given action within the system.

    """

    def __init__(self, user_id=None, project_id=None, is_admin=None,
                 read_deleted="no", remote_address=None, timestamp=None,
                 quota_class=None, service_catalog=None,
                 user_auth_plugin=None, **kwargs):
        """:param read_deleted: 'no' indicates deleted records are hidden,
                'yes' indicates deleted records are visible,
                'only' indicates that *only* deleted records are visible.

           :param overwrite: Set to False to ensure that the greenthread local
                copy of the index is not overwritten.

           :param instance_lock_checked: This is not used and will be removed
                in a future release.

           :param user_auth_plugin: The auth plugin for the current request's
                authentication data.
        """
        if user_id:
            kwargs['user_id'] = user_id
        if project_id:
            kwargs['project_id'] = project_id

        super(RequestContext, self).__init__(is_admin=is_admin, **kwargs)

        self.read_deleted = read_deleted
        self.remote_address = remote_address
        if not timestamp:
            timestamp = timeutils.utcnow()
        if isinstance(timestamp, str):
            timestamp = timeutils.parse_strtime(timestamp)
        self.timestamp = timestamp

        if service_catalog:
            # Only include required parts of service_catalog
            self.service_catalog = [s for s in service_catalog
                                    if s.get('type') in ('image')]
        else:
            # if list is empty or none
            self.service_catalog = []

        self.user_auth_plugin = user_auth_plugin
        # if self.is_admin is None:
        #    self.is_admin = policy.check_is_admin(self)

    def get_auth_plugin(self):
        if self.user_auth_plugin:
            return self.user_auth_plugin
        else:
            return _ContextAuthPlugin(self.auth_token, self.service_catalog)

    def to_dict(self):
        values = super(RequestContext, self).to_dict()
        # FIXME(dims): defensive hasattr() checks need to be
        # removed once we figure out why we are seeing stack
        # traces
        values.update({
            'user_id': getattr(self, 'user_id', None),
            'project_id': getattr(self, 'project_id', None),
            'is_admin': getattr(self, 'is_admin', None),
            'read_deleted': getattr(self, 'read_deleted', 'no'),
            'remote_address': getattr(self, 'remote_address', None),
            'timestamp': utils.strtime(self.timestamp) if hasattr(
                self, 'timestamp') else None,
            'request_id': getattr(self, 'request_id', None),
            'quota_class': getattr(self, 'quota_class', None),
            'user_name': getattr(self, 'user_name', None),
            'service_catalog': getattr(self, 'service_catalog', None),
            'project_name': getattr(self, 'project_name', None),
        })
        # NOTE(tonyb): This can be removed once we're certain to have a
        # RequestContext contains 'is_admin_project', We can only get away with
        # this because we "know" the default value of 'is_admin_project' which
        # is very fragile.
        values.update({
            'is_admin_project': getattr(self, 'is_admin_project', True),
        })
        return values

    @classmethod
    def from_dict(cls, values):
        return super(RequestContext, cls).from_dict(
            values,
            user_id=values.get('user_id'),
            project_id=values.get('project_id'),
            # TODO(sdague): oslo.context has show_deleted, if
            # possible, we should migrate to that in the future so we
            # don't need to be different here.
            read_deleted=values.get('read_deleted', 'no'),
            remote_address=values.get('remote_address'),
            timestamp=values.get('timestamp'),
            quota_class=values.get('quota_class'),
            service_catalog=values.get('service_catalog'),
        )


def get_context():
    """A helper method to get a blank context.

    Note that overwrite is False here so this context will not update the
    greenthread-local stored context that is used when logging.
    """
    return RequestContext(user_id=None,
                          project_id=None,
                          is_admin=False,
                          overwrite=False)


def get_admin_context(read_deleted="no"):
    # NOTE(alaski): This method should only be used when an admin context is
    # necessary for the entirety of the context lifetime. If that's not the
    # case please use get_context(), or create the RequestContext manually, and
    # use context.elevated() where necessary. Some periodic tasks may use
    # get_admin_context so that their database calls are not filtered on
    # project_id.
    return RequestContext(user_id=None,
                          project_id=None,
                          is_admin=True,
                          read_deleted=read_deleted,
                          overwrite=False)


def is_user_context(context):
    """Indicates if the request context is a normal user."""
    if not context:
        return False
    if context.is_admin:
        return False
    if not context.user_id or not context.project_id:
        return False
    return True


def require_context(ctxt):
    """Raise exception.Forbidden() if context is not a user or an
    admin context.
    """
    if not ctxt.is_admin and not is_user_context(ctxt):
        raise exception.Forbidden()


def authorize_project_context(context, project_id):
    """Ensures a request has permission to access the given project."""
    if is_user_context(context):
        if not context.project_id:
            raise exception.Forbidden()
        elif context.project_id != project_id:
            raise exception.Forbidden()


def authorize_user_context(context, user_id):
    """Ensures a request has permission to access the given user."""
    if is_user_context(context):
        if not context.user_id:
            raise exception.Forbidden()
        elif context.user_id != user_id:
            raise exception.Forbidden()



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_serialization import jsonutils
from pecan import hooks

from cyborg.common import exception
from cyborg.conductor import rpcapi
from cyborg import context


class ConfigHook(hooks.PecanHook):
    """Attach the config object to the request so controllers can get to it."""

    def before(self, state):
        state.request.cfg = cfg.CONF


class PublicUrlHook(hooks.PecanHook):
    """Attach the right public_url to the request.

    Attach the right public_url to the request so resources can create
    links even when the API service is behind a proxy or SSL terminator.
    """

    def before(self, state):
        state.request.public_url = (
            cfg.CONF.api.public_endpoint or state.request.host_url)


class ConductorAPIHook(hooks.PecanHook):
    """Attach the conductor_api object to the request."""

    def __init__(self):
        self.conductor_api = rpcapi.ConductorAPI()

    def before(self, state):
        state.request.conductor_api = self.conductor_api


class ContextHook(hooks.PecanHook):
    """Configures a request context and attaches it to the request.

    The following HTTP request headers are used:

    X-User-Id or X-User:
        Used for context.user_id.

    X-Tenant-Id or X-Tenant:
        Used for context.project_id.

    X-Auth-Token:
        Used for context.auth_token.

    X-Roles:
        Used for setting context.is_admin flag to either True or False.
        The flag is set to True, if X-Roles contains either an administrator
        or admin substring. Otherwise it is set to False.

    """

    def __init__(self, public_api_routes):
        self.public_api_routes = public_api_routes
        super(ContextHook, self).__init__()

    def before(self, state):
        req = state.request

        service_catalog = None
        if req.headers.get('X_SERVICE_CATALOG') is not None:
            try:
                catalog_header = req.headers.get('X_SERVICE_CATALOG')
                service_catalog = jsonutils.loads(catalog_header)
            except ValueError:
                raise exception.InvalidJsonType('service catalog')

        # NOTE(jamielennox): This is a full auth plugin set by auth_token
        # middleware in newer versions.
        user_auth_plugin = req.environ.get('keystone.token_auth')

        roles = req.headers.get('X-Roles', '').split(',')
        is_admin = ('admin' in roles or 'administrator' in roles)

        state.request.context = context.RequestContext.from_environ(
            req.environ,
            user_auth_plugin=user_auth_plugin,
            is_admin=is_admin,
            service_catalog=service_catalog)



# Copyright 2018 Beijing Lenovo Software Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys
from unittest import mock

from oslo_serialization import jsonutils

import cyborg
from cyborg.accelerator.drivers.gpu.nvidia.driver import NVIDIAGPUDriver
from cyborg.accelerator.drivers.gpu import utils
from cyborg.tests import base


CONF = cyborg.conf.CONF

NVIDIA_GPU_INFO = "0000:00:06.0 3D controller [0302]: NVIDIA Corporation " \
                  "GP100GL [Tesla P100 PCIe 12GB] [10de:15f7] (rev a1)"

NVIDIA_T4_GPU_INFO = "0000:af:00.0 3D controller [0302]: NVIDIA Corporation "\
                     "TU104GL [Tesla T4] [10de:1eb8] (rev a1)"

NVIDIA_T4_SUPPORTED_MDEV_TYPES = ['nvidia-222', 'nvidia-223', 'nvidia-224',
                                  'nvidia-225', 'nvidia-226', 'nvidia-227',
                                  'nvidia-228', 'nvidia-229', 'nvidia-230',
                                  'nvidia-231', 'nvidia-232', 'nvidia-233',
                                  'nvidia-234', 'nvidia-252', 'nvidia-319',
                                  'nvidia-320', 'nvidia-321']

BUILTIN = '__builtin__' if (sys.version_info[0] < 3) else '__builtins__'


class stdout(object):
    def readlines(self):
        return [NVIDIA_GPU_INFO]

    def readlines_T4(self):
        return [NVIDIA_T4_GPU_INFO]


class p(object):
    def __init__(self):
        self.stdout = stdout()

    def wait(self):
        pass


class TestGPUDriverUtils(base.TestCase):

    def setUp(self):
        super(TestGPUDriverUtils, self).setUp()
        self.p = p()

    @mock.patch('cyborg.accelerator.drivers.gpu.utils.lspci_privileged')
    def test_discover_vendors(self, mock_devices):
        mock_devices.return_value = self.p.stdout.readlines()
        gpu_vendors = utils.discover_vendors()
        self.assertEqual(1, len(gpu_vendors))

    @mock.patch('cyborg.accelerator.drivers.gpu.utils.lspci_privileged')
    def test_discover_gpus_report_pGPU(self, mock_devices_for_vendor):
        """test nvidia pGPU discover"""
        mock_devices_for_vendor.return_value = self.p.stdout.readlines()
        self.set_defaults(host='host-192-168-32-195', debug=True)

        nvidia = NVIDIAGPUDriver()
        gpu_list = nvidia.discover()

        self.assertEqual(1, len(gpu_list))
        attach_handle_list = [
            {'attach_type': 'PCI',
             'attach_info': '{"bus": "00", '
                            '"device": "06", '
                            '"domain": "0000", '
                            '"function": "0"}',
             'in_use': False}
        ]
        attribute_list = [
            {'key': 'rc', 'value': 'PGPU'},
            {'key': 'trait0', 'value': 'OWNER_CYBORG'},
            {'key': 'trait1', 'value': 'CUSTOM_NVIDIA_15F7'},
        ]
        expected = {
            'vendor': '10de',
            'type': 'GPU',
            'model': 'NVIDIA Corporation GP100GL [Tesla P100 PCIe 12GB]',
            'std_board_info':
                {"controller": "3D controller", "product_id": "15f7"},
            'vendor_board_info': {"vendor_info": "gpu_vb_info"},
            'deployable_list':
                [
                    {
                        'num_accelerators': 1,
                        'driver_name': 'NVIDIA',
                        'name': 'host-192-168-32-195_0000:00:06.0',
                        'attach_handle_list': attach_handle_list,
                        'attribute_list': attribute_list
                    },
                ],
            'controlpath_id': {'cpid_info': '{"bus": "00", '
                                            '"device": "06", '
                                            '"domain": "0000", '
                                            '"function": "0"}',
                               'cpid_type': 'PCI'}
            }
        gpu_obj = gpu_list[0]
        gpu_dict = gpu_obj.as_dict()
        gpu_dep_list = gpu_dict['deployable_list']
        gpu_attach_handle_list = \
            gpu_dep_list[0].as_dict()['attach_handle_list']
        gpu_attribute_list = \
            gpu_dep_list[0].as_dict()['attribute_list']
        attri_obj_data = []
        [attri_obj_data.append(attr.as_dict()) for attr in gpu_attribute_list]
        attribute_actual_data = sorted(attri_obj_data, key=lambda i: i['key'])
        self.assertEqual(expected['vendor'], gpu_dict['vendor'])
        self.assertEqual(expected['model'], gpu_dict['model'])
        self.assertEqual(expected['controlpath_id'],
                         gpu_dict['controlpath_id'])
        self.assertEqual(expected['std_board_info'],
                         jsonutils.loads(gpu_dict['std_board_info']))
        self.assertEqual(expected['vendor_board_info'],
                         jsonutils.loads(gpu_dict['vendor_board_info']))
        self.assertEqual(expected['deployable_list'][0]['num_accelerators'],
                         gpu_dep_list[0].as_dict()['num_accelerators'])
        self.assertEqual(expected['deployable_list'][0]['name'],
                         gpu_dep_list[0].as_dict()['name'])
        self.assertEqual(expected['deployable_list'][0]['driver_name'],
                         gpu_dep_list[0].as_dict()['driver_name'])
        self.assertEqual(attach_handle_list[0],
                         gpu_attach_handle_list[0].as_dict())
        self.assertEqual(attribute_list, attribute_actual_data)

    @mock.patch('builtins.open')
    @mock.patch('os.listdir')
    @mock.patch('os.path.exists')
    @mock.patch('cyborg.accelerator.drivers.gpu.utils.lspci_privileged')
    def test_discover_gpus_report_vGPU(self, mock_devices_for_vendor,
                                       mock_path_exists,
                                       mock_supported_mdev_types,
                                       mock_open):
        """test nvidia vGPU discover"""
        mock_devices_for_vendor.return_value = self.p.stdout.readlines_T4()
        mock_path_exists.return_value = True
        mock_supported_mdev_types.return_value = NVIDIA_T4_SUPPORTED_MDEV_TYPES
        file_content_list = ['GRID T4-1B', '1']
        mock_open.side_effect = multi_mock_open(*file_content_list)
        self.set_defaults(host='host-192-168-32-195', debug=True)
        self.set_defaults(enabled_vgpu_types='nvidia-223', group='gpu_devices')
        cyborg.conf.devices.register_dynamic_opts(CONF)
        self.set_defaults(
            device_addresses=['0000:af:00.0'], group='vgpu_nvidia-223')
        nvidia = NVIDIAGPUDriver()
        gpu_list = nvidia.discover()

        self.assertEqual(1, len(gpu_list))
        attach_handle_list = [
            {'attach_type': 'MDEV',
             'attach_info': '{"asked_type": "nvidia-223", '
                            '"bus": "af", '
                            '"device": "00", '
                            '"domain": "0000", '
                            '"function": "0", '
                            '"vgpu_mark": "nvidia-223_0"}',
             'in_use': False}
        ] * 8
        attribute_list = [
            {'key': 'rc', 'value': 'VGPU'},
            {'key': 'trait0', 'value': 'OWNER_CYBORG'},
            {'key': 'trait1', 'value': 'CUSTOM_NVIDIA_1EB8_T4_1B'},
        ]
        expected = {
            'vendor': '10de',
            'type': 'GPU',
            'std_board_info':
                {"controller": "3D controller", "product_id": "1eb8"},
            'vendor_board_info': {"vendor_info": "gpu_vb_info"},
            'deployable_list':
                [
                    {
                        'num_accelerators': 18,
                        'driver_name': 'NVIDIA',
                        'name': 'host-192-168-32-195_0000:af:00.0',
                        'attach_handle_list': attach_handle_list,
                        'attribute_list': attribute_list
                    },
                ],
            'controlpath_id': {'cpid_info': '{"bus": "af", '
                                            '"device": "00", '
                                            '"domain": "0000", '
                                            '"function": "0"}',
                               'cpid_type': 'PCI'}
            }
        gpu_obj = gpu_list[0]
        gpu_dict = gpu_obj.as_dict()
        gpu_dep_list = gpu_dict['deployable_list']
        gpu_attach_handle_list = gpu_dep_list[0].as_dict()[
            'attach_handle_list']
        gpu_attribute_list = gpu_dep_list[0].as_dict()['attribute_list']
        attri_obj_data = []
        [attri_obj_data.append(attr.as_dict()) for attr in gpu_attribute_list]
        attribute_actual_data = sorted(attri_obj_data, key=lambda i: i['key'])
        self.assertEqual(expected['vendor'], gpu_dict['vendor'])
        self.assertEqual(expected['controlpath_id'],
                         gpu_dict['controlpath_id'])
        self.assertEqual(expected['std_board_info'],
                         jsonutils.loads(gpu_dict['std_board_info']))
        self.assertEqual(expected['vendor_board_info'],
                         jsonutils.loads(gpu_dict['vendor_board_info']))
        self.assertEqual(expected['deployable_list'][0]['num_accelerators'],
                         gpu_dep_list[0].as_dict()['num_accelerators'])
        self.assertEqual(expected['deployable_list'][0]['name'],
                         gpu_dep_list[0].as_dict()['name'])
        self.assertEqual(expected['deployable_list'][0]['driver_name'],
                         gpu_dep_list[0].as_dict()['driver_name'])
        self.assertEqual(attach_handle_list[0],
                         gpu_attach_handle_list[0].as_dict())
        self.assertEqual(attribute_list, attribute_actual_data)


def multi_mock_open(*file_contents):
    """Create a mock "open" that will mock open multiple files in sequence.

    : params file_contents:  a list of file contents to be returned by open

    : returns: (MagicMock) a mock opener that will return the contents of the
               first file when opened the first time, the second file when
               opened the second time, etc.
    """

    mock_files = [
        mock.mock_open(read_data=content).return_value for content in
        file_contents]
    mock_opener = mock.mock_open()
    mock_opener.side_effect = mock_files
    return mock_opener



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add-quota-related-tables

Revision ID: d6f033d8fa5b
Revises: f50980397351
Create Date: 2018-04-28 03:07:06.857245

"""

# revision identifiers, used by Alembic.
revision = 'd6f033d8fa5b'
down_revision = 'f50980397351'

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'quota_usages',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('project_id', sa.String(length=255), nullable=True),
        sa.Column('user_id', sa.String(length=255), nullable=True),
        sa.Column('resource', sa.String(length=255), nullable=False),
        sa.Column('in_use', sa.Integer(), nullable=False),
        sa.Column('reserved', sa.Integer(), nullable=False),
        sa.Column('until_refresh', sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_index('ix_quota_usages_project_id', 'quota_usages',
                    ['project_id'], unique=False)
    op.create_index('ix_quota_usages_user_id', 'quota_usages', ['user_id'],
                    unique=False)

    op.create_table(
        'reservations',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=False),
        sa.Column('usage_id', sa.Integer(), nullable=False),
        sa.Column('project_id', sa.String(length=255), nullable=True),
        sa.Column('user_id', sa.String(length=255), nullable=True),
        sa.Column('resource', sa.String(length=255), nullable=True),
        sa.Column('delta', sa.Integer(), nullable=False),
        sa.Column('expire', sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(['usage_id'], ['quota_usages.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_index('ix_reservations_project_id', 'reservations',
                    ['project_id'], unique=False)
    op.create_index('ix_reservations_user_id', 'reservations', ['user_id'],
                    unique=False)
    op.create_index('reservations_uuid_idx', 'reservations', ['uuid'],
                    unique=False)
    # ### end Alembic commands ###



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
SPDK VHOSTDRIVER module implementation.
"""

from oslo_log import log as logging

from cyborg.accelerator.drivers.spdk.spdk import SPDKDRIVER
from cyborg.accelerator.drivers.spdk.util import common_fun
from cyborg.accelerator.drivers.spdk.util.pyspdk.py_spdk import PySPDK
from cyborg.accelerator.drivers.spdk.util.pyspdk.vhost_client import VhostTgt

LOG = logging.getLogger(__name__)


class VHOSTDRIVER(SPDKDRIVER):
    """VHOSTDRIVER class.

       vhost server app should be able to implement this driver.
    """

    SERVER = 'vhost'

    def __init__(self, *args, **kwargs):
        super(VHOSTDRIVER, self).__init__(*args, **kwargs)
        self.servers = common_fun.discover_servers()
        self.py = common_fun.get_py_client(self.SERVER)

    def discover_accelerator(self):
        if common_fun.check_for_setup_error(self.py, self.SERVER):
            return self.get_one_accelerator()

    def get_one_accelerator(self):
        acc_client = VhostTgt(self.py)
        bdevs = acc_client.get_bdevs()
        # Display current blockdev list
        scsi_devices = acc_client.get_scsi_devices()
        # Display SCSI devices
        luns = acc_client.get_luns()
        # Display active LUNs
        interfaces = acc_client.get_interfaces()
        # Display current interface list
        accelerator_obj = {
            'server': self.SERVER,
            'bdevs': bdevs,
            'scsi_devices': scsi_devices,
            'luns': luns,
            'interfaces': interfaces
        }
        return accelerator_obj

    def install_accelerator(self, driver_id, driver_type):
        pass

    def uninstall_accelerator(self, driver_id, driver_type):
        pass

    def accelerator_list(self):
        return self.get_all_accelerators()

    def get_all_accelerators(self):
        accelerators = []
        for accelerator_i in range(len(self.servers)):
            accelerator = self.servers[accelerator_i]
            py_tmp = PySPDK(accelerator)
            if py_tmp.is_alive():
                accelerators.append(self.get_one_accelerator())
        return accelerators

    def update(self, driver_type, **kwargs):
        pass

    def attach_instance(self, instance_id):
        pass

    def detach_instance(self, instance_id):
        pass

    def add_ip_address(self, ifc_index, ip_addr):
        """Add IP address

        :param ifc_index: ifc index of the nic device.
        :param ip_addr: ip address will be added.
        :return: ip_address
        """
        acc_client = VhostTgt(self.py)
        return acc_client.add_ip_address(ifc_index, ip_addr)

    def delete_ip_address(self, ifc_index, ip_addr):
        """Delete IP address

        :param ifc_index: ifc index of the nic device.
        :param ip_addr: ip address will be added.
        :return: ip_address
        """
        acc_client = VhostTgt(self.py)
        return acc_client.delete_ip_address(ifc_index, ip_addr)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cyborg.accelerator.drivers.driver import GenericDriver
from cyborg.tests import base


class WellDoneDriver(GenericDriver):
    def discover(self):
        pass

    def update(self, control_path, image_path):
        pass

    def get_stats(self):
        pass


class NotCompleteDriver(GenericDriver):
    def discover(self):
        pass


class TestGenericDriver(base.TestCase):

    def test_generic_driver(self):
        # Can't instantiate abstract class NotCompleteDriver with
        # abstract methods get_stats, update
        result = self.assertRaises(TypeError, NotCompleteDriver)
        self.assertIn("Can't instantiate abstract class",
                      str(result))



# Copyright 2018 Lenovo (Beijing) Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import base as object_base

from cyborg.objects import base
from cyborg.objects.control_path import ControlpathID
from cyborg.objects import fields as object_fields


@base.CyborgObjectRegistry.register
class DriverControlPathID(base.DriverObjectBase,
                          object_base.VersionedObjectDictCompat):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'cpid_type': object_fields.StringField(nullable=False),
        # PCI BDF, PowerVM device, etc.
        'cpid_info': object_fields.StringField(nullable=False)
    }

    def create(self, context, device_id):
        """Create a driver-side ControlPathID for drivers. Call
        ControlpathID object to store in DB.
        """
        cpid_obj = ControlpathID(context=context,
                                 device_id=device_id,
                                 cpid_type=self.cpid_type,
                                 cpid_info=self.cpid_info)
        cpid_obj.create(context)
        return cpid_obj

    def destroy(self, context, device_id):
        cpid_obj = ControlpathID.get_by_device_id_cpidinfo(context,
                                                           device_id,
                                                           self.cpid_info)
        if cpid_obj is not None:
            cpid_obj.destroy(context)

    @classmethod
    def get(cls, context, device_id):
        # return None when can't found any.
        cpid_obj = ControlpathID.get_by_device_id(context, device_id)
        driver_cpid_obj = None
        if cpid_obj is not None:
            driver_cpid_obj = cls(context=context,
                                  cpid_type=cpid_obj.cpid_type,
                                  cpid_info=cpid_obj.cpid_info)
        return driver_cpid_obj



# Copyright 2018 Huawei Technologies Co.,LTD.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from oslo_log import log

from oslo_context import context

from cyborg.objects import base
from cyborg.objects import fields
from cyborg import tests as test


LOG = log.getLogger(__name__)


class MyOwnedObject(base.CyborgPersistentObject, base.CyborgObject):
    VERSION = '1.0'
    fields = {'baz': fields.IntegerField()}


class MyObj(base.CyborgPersistentObject, base.CyborgObject,
            base.CyborgObjectDictCompat):
    VERSION = '1.6'
    fields = {'foo': fields.IntegerField(default=1),
              'bar': fields.StringField(),
              'missing': fields.StringField(),
              'readonly': fields.IntegerField(read_only=True),
              'rel_object': fields.ObjectField('MyOwnedObject', nullable=True),
              'rel_objects': fields.ListOfObjectsField('MyOwnedObject',
                                                       nullable=True),
              'mutable_default': fields.ListOfStringsField(default=[]),
              }

    @staticmethod
    def _from_db_object(context, obj, db_obj):
        self = MyObj()
        self.foo = db_obj['foo']
        self.bar = db_obj['bar']
        self.missing = db_obj['missing']
        self.readonly = 1
        self._context = context
        return self

    def obj_load_attr(self, attrname):
        setattr(self, attrname, 'loaded!')

    def query(cls, context):
        obj = cls(context=context, foo=1, bar='bar')
        obj.obj_reset_changes()
        return obj

    def marco(self):
        return 'polo'

    def _update_test(self):
        self.bar = 'updated'

    def save(self):
        self.obj_reset_changes()

    def refresh(self):
        self.foo = 321
        self.bar = 'refreshed'
        self.obj_reset_changes()

    def modify_save_modify(self):
        self.bar = 'meow'
        self.save()
        self.foo = 42
        self.rel_object = MyOwnedObject(baz=42)

    def obj_make_compatible(self, primitive, target_version):
        super(MyObj, self).obj_make_compatible(primitive, target_version)
        # NOTE(danms): Simulate an older version that had a different
        # format for the 'bar' attribute
        if target_version == '1.1' and 'bar' in primitive:
            primitive['bar'] = 'old%s' % primitive['bar']


class RandomMixInWithNoFields(object):
    """Used to test object inheritance using a mixin that has no fields."""
    pass


@base.CyborgObjectRegistry.register_if(False)
class TestSubclassedObject(RandomMixInWithNoFields, MyObj):
    fields = {'new_field': fields.StringField()}


class TestObjToPrimitive(test.base.TestCase):

    def test_obj_to_primitive_list(self):
        @base.CyborgObjectRegistry.register_if(False)
        class MyObjElement(base.CyborgObject):
            fields = {'foo': fields.IntegerField()}

            def __init__(self, foo):
                super(MyObjElement, self).__init__()
                self.foo = foo

        @base.CyborgObjectRegistry.register_if(False)
        class MyList(base.ObjectListBase, base.CyborgObject):
            fields = {'objects': fields.ListOfObjectsField('MyObjElement')}

        mylist = MyList()
        mylist.objects = [MyObjElement(1), MyObjElement(2), MyObjElement(3)]
        self.assertEqual([1, 2, 3],
                         [x['foo'] for x in base.obj_to_primitive(mylist)])

    def test_obj_to_primitive_dict(self):
        base.CyborgObjectRegistry.register(MyObj)
        myobj = MyObj(foo=1, bar='foo')
        self.assertEqual({'foo': 1, 'bar': 'foo'},
                         base.obj_to_primitive(myobj))

    def test_obj_to_primitive_recursive(self):
        base.CyborgObjectRegistry.register(MyObj)

        class MyList(base.ObjectListBase, base.CyborgObject):
            fields = {'objects': fields.ListOfObjectsField('MyObj')}

        mylist = MyList(objects=[MyObj(), MyObj()])
        for i, value in enumerate(mylist):
            value.foo = i
        self.assertEqual([{'foo': 0}, {'foo': 1}],
                         base.obj_to_primitive(mylist))

    def test_obj_to_primitive_with_ip_addr(self):
        @base.CyborgObjectRegistry.register_if(False)
        class TestObject(base.CyborgObject):
            fields = {'addr': fields.IPAddressField(),
                      'cidr': fields.IPNetworkField()}

        obj = TestObject(addr='1.2.3.4', cidr='1.1.1.1/16')
        self.assertEqual({'addr': '1.2.3.4', 'cidr': '1.1.1.1/16'},
                         base.obj_to_primitive(obj))


def compare_obj(test, obj, db_obj, subs=None, allow_missing=None,
                comparators=None):
    """Compare a CyborgObject and a dict-like database object.

    This automatically converts TZ-aware datetimes and iterates over
    the fields of the object.

    :param:test: The TestCase doing the comparison
    :param:obj: The CyborgObject to examine
    :param:db_obj: The dict-like database object to use as reference
    :param:subs: A dict of objkey=dbkey field substitutions
    :param:allow_missing: A list of fields that may not be in db_obj
    :param:comparators: Map of comparator functions to use for certain fields
    """

    if subs is None:
        subs = {}
    if allow_missing is None:
        allow_missing = []
    if comparators is None:
        comparators = {}

    for key in obj.fields:
        if key in allow_missing and not obj.obj_attr_is_set(key):
            continue
        obj_val = getattr(obj, key)
        db_key = subs.get(key, key)
        db_val = db_obj[db_key]
        if isinstance(obj_val, datetime.datetime):
            obj_val = obj_val.replace(tzinfo=None)

        if key in comparators:
            comparator = comparators[key]
            comparator(db_val, obj_val)
        else:
            test.assertEqual(db_val, obj_val)


class _BaseTestCase(test.base.TestCase):
    def setUp(self):
        super(_BaseTestCase, self).setUp()
        self.user_id = 'fake-user'
        self.project_id = 'fake-project'
        self.context = context.RequestContext(self.user_id, self.project_id)

        base.CyborgObjectRegistry.register(MyObj)
        base.CyborgObjectRegistry.register(MyOwnedObject)

    def compare_obj(self, obj, db_obj, subs=None, allow_missing=None,
                    comparators=None):
        compare_obj(self, obj, db_obj, subs=subs, allow_missing=allow_missing,
                    comparators=comparators)

    def str_comparator(self, expected, obj_val):
        """Compare an object field to a string in the db by performing
        a simple coercion on the object field value.
        """
        self.assertEqual(expected, str(obj_val))


class _LocalTest(_BaseTestCase):
    def setUp(self):
        super(_LocalTest, self).setUp()



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Cyborg Default Config Setting"""

import os
import socket

from oslo_config import cfg

from cyborg.common.i18n import _


exc_log_opts = [
    cfg.BoolOpt('fatal_exception_format_errors',
                default=False,
                help=_('Used if there is a formatting error when generating '
                       'an exception message (a programming error). If True, '
                       'raise an exception; if False, use the unformatted '
                       'message.')),
]

service_opts = [
    cfg.HostAddressOpt('host',
                       default=socket.gethostname(),
                       sample_default='localhost',
                       help=_('Name of this node. This can be an opaque '
                              'identifier. It is not necessarily a hostname, '
                              'FQDN, or IP address. However, the node name '
                              'must be valid within an AMQP key, and if using '
                              'ZeroMQ, a valid hostname, FQDN, or IP address.')
                       ),
    cfg.IntOpt('periodic_interval',
               default=60,
               help=_('Default interval (in seconds) for running periodic '
                      'tasks.')),
    cfg.IntOpt(
        'thread_pool_size',
        default=10,
        help=_('This option specifies the size of the pool of threads used '
               'by API to do async jobs.It is possible to limit the number '
               'of concurrent connections using this option.')),
    cfg.IntOpt(
        'bind_timeout',
        default=60,
        help=_('This option specifies the timeout of async job for ARQ '
               'bind.')),
]

path_opts = [
    cfg.StrOpt('pybasedir',
               default=os.path.abspath(
                   os.path.join(os.path.dirname(__file__), '../')),
               sample_default='/usr/lib/python/site-packages/cyborg/cyborg',
               help=_('Directory where the cyborg python module is '
                      'installed.')),
    cfg.StrOpt('bindir',
               default='$pybasedir/bin',
               help=_('Directory where cyborg binaries are installed.')),
    cfg.StrOpt('state_path',
               default='$pybasedir',
               help=_("Top-level directory for maintaining cyborg's state.")),
]


def register_opts(conf):
    conf.register_opts(exc_log_opts)
    conf.register_opts(service_opts)
    conf.register_opts(path_opts)


DEFAULT_OPTS = (exc_log_opts + service_opts + path_opts)


def list_opts():
    return {
        'DEFAULT': DEFAULT_OPTS
    }



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import functools

import inspect

import pecan
from pecan import rest
from webob import exc
import wsme
from wsme import types as wtypes

API_V2 = 'v2'
# name of attribute to keep version method information


class APIBase(wtypes.Base):
    created_at = wsme.wsattr(datetime.datetime, readonly=True)
    """The time in UTC at which the object is created"""

    updated_at = wsme.wsattr(datetime.datetime, readonly=True)
    """The time in UTC at which the object is updated"""

    def as_dict(self):
        """Render this object as a dict of its fields."""
        return {k: getattr(self, k) for k in self.fields
                if hasattr(self, k) and getattr(self, k) != wsme.Unset}


class CyborgController(rest.RestController):

    def _handle_patch(self, method, remainder, request=None):
        """Routes ``PATCH`` _custom_actions."""
        # route to a patch_all or get if no additional parts are available
        if not remainder or remainder == ['']:
            controller = self._find_controller('patch_all', 'patch')
            if controller:
                return controller, []
            pecan.abort(404)

        controller = getattr(self, remainder[0], None)
        if controller and not inspect.ismethod(controller):
            return pecan.routing.lookup_controller(controller, remainder[1:])
        # route to custom_action
        match = self._handle_custom_action(method, remainder, request)
        if match:
            return match

        # finally, check for the regular patch_one/patch requests
        controller = self._find_controller('patch_one', 'patch')
        if controller:
            return controller, remainder

        pecan.abort(405)


@functools.total_ordering
class Version(object):
    """API Version object."""

    current_api_version = 'OpenStack-API-Version'
    """HTTP Header string carrying the requested version"""

    min_api_version = 'OpenStack-API-Minimum-Version'
    """HTTP response header"""

    max_api_version = 'OpenStack-API-Maximum-Version'
    """HTTP response header"""

    def __init__(self, headers, default_version, latest_version):
        """Create an API Version object from the supplied headers.

        :param headers: webob headers
        :param default_version: version to use if not specified in headers
        :param latest_version: version to use if latest is requested
        :raises: webob.HTTPNotAcceptable

        """
        (self.major, self.minor) = Version.parse_headers(
            headers, default_version, latest_version)

    def __repr__(self):
        return '%s.%s' % (self.major, self.minor)

    @staticmethod
    def parse_headers(headers, default_version, latest_version):
        """Determine the API version requested based on the headers supplied.

        :param headers: webob headers
        :param default_version: version to use if not specified in headers
        :param latest_version: version to use if latest is requested
        :returns: a tuple of (major, minor) version numbers
        :raises: webob.HTTPNotAcceptable

        """
        version_str = headers.get(Version.current_api_version, default_version)
        minimal_version = (2, 0)

        if version_str is None:
            # If requested header is wrong, Cyborg answers with the minimal
            # supported version.
            return minimal_version

        if version_str.lower() == 'latest':
            parse_str = latest_version
        else:
            parse_str = version_str

        try:
            version = tuple(int(i) for i in parse_str.split('.'))
        except ValueError:
            version = minimal_version

        if len(version) != 2:
            raise exc.HTTPNotAcceptable(
                "Invalid value for %s header" % Version.current_api_version)
        return version

    def __gt__(self, other):
        return (self.major, self.minor) > (other.major, other.minor)

    def __eq__(self, other):
        return (self.major, self.minor) == (other.major, other.minor)

    def __ne__(self, other):
        return not self.__eq__(other)



# Copyright 2022 Inspur.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import HTTPStatus
import pecan
import wsme
from wsme import types as wtypes

from oslo_log import log

from cyborg.api.controllers import base
from cyborg.api.controllers import link
from cyborg.api.controllers import types
from cyborg.api import expose
from cyborg.common import authorize_wsgi
from cyborg import objects
LOG = log.getLogger(__name__)


class Attribute(base.APIBase):
    """API representation of a attribute.

    This class enforces type checking and value constraints, and converts
    between the internal object model and the API representation of
    a attribute. See module notes above.
    """

    """The UUID of the attribute"""
    uuid = types.uuid

    """The deployable_id of the attribute"""
    deployable_id = wtypes.IntegerType()

    """The key of the device attribute"""
    key = wtypes.text

    """The value of the attribute"""
    value = wtypes.text

    created_at = wtypes.datetime.datetime
    updated_at = wtypes.datetime.datetime

    """A list containing a self link"""
    links = wsme.wsattr([link.Link], readonly=True)

    def __init__(self, **kwargs):
        super(Attribute, self).__init__(**kwargs)
        self.fields = []
        for field in objects.Attribute.fields:
            self.fields.append(field)
            setattr(self, field, kwargs.get(field, wtypes.Unset))

    @classmethod
    def convert_with_links(cls, obj_attribute):
        api_attribute = cls(**obj_attribute.as_dict())
        api_attribute.links = [
            link.Link.make_link('self', pecan.request.public_url,
                                'attributes', api_attribute.uuid)
            ]
        return api_attribute

    def get_attribute(self, obj_attribute):
        api_obj = {}
        for field in ['uuid', 'deployable_id', 'key', 'value']:
            api_obj[field] = obj_attribute[field]
        for field in ['created_at', 'updated_at']:
            api_obj[field] = str(obj_attribute[field])
        api_obj['links'] = [
            link.Link.make_link_dict('attributes', api_obj['uuid'])
            ]
        return api_obj


class AttributeCollection(Attribute):
    """API representation of a collection of attributes."""

    """A list containing attribute objects"""
    attributes = [Attribute]

    @classmethod
    def convert_with_links(cls, obj_attributes):
        collection = cls()
        collection.attributes = [
            Attribute.convert_with_links(obj_attribute)
            for obj_attribute in obj_attributes]
        return collection

    def get_attributes(self, obj_attributes):
        api_obj_attributes = [
            self.get_attribute(obj_attribute)
            for obj_attribute in obj_attributes]
        return api_obj_attributes


class AttributesController(base.CyborgController,
                           AttributeCollection):
    """REST controller for Attributes."""

    @authorize_wsgi.authorize_wsgi("cyborg:attribute", "get_all", False)
    @expose.expose(AttributeCollection, wtypes.IntegerType(), wtypes.text)
    def get_all(self, deployable_id=None, key=None):
        """Retrieve a list of attributes."""
        LOG.info('[attributes] get_all by deployable_id:(%s) and key:(%s).',
                 deployable_id, key)
        search_opts = {}
        if deployable_id:
            search_opts['deployable_id'] = deployable_id
        if key:
            search_opts['key'] = key
        context = pecan.request.context
        attributes = objects.Attribute.get_by_filter(context, search_opts)
        ret = AttributeCollection.convert_with_links(attributes)
        LOG.info('[Attributes] get_all returned: %s', ret)
        return ret

    @authorize_wsgi.authorize_wsgi("cyborg:attribute", "get_one")
    @expose.expose(Attribute, types.uuid)
    def get_one(self, uuid):
        """Retrieve a single attribute by uuid."""
        LOG.info('[attributes] get by uuid: %s.', uuid)
        context = pecan.request.context
        api_obj_attribute = objects.Attribute.get(context, uuid)
        ret = Attribute.convert_with_links(api_obj_attribute)
        LOG.info('[attributes] get_one returned: %s', ret)
        return ret

    @authorize_wsgi.authorize_wsgi("cyborg:attribute", "create", False)
    @expose.expose(Attribute, body=types.jsontype,
                   status_code=HTTPStatus.CREATED)
    def post(self, req_attr):
        """Create one attribute.
        :param req_attr: attribute value.
         { "deployable_id": <integer>,
           "key": <string>,
           "value": <string>
         }
         :returns: The object of created attribute
        """
        LOG.info("[attributes] POST request = (%s)", req_attr)
        context = pecan.request.context
        attribute = objects.Attribute(context, **req_attr).create(context)
        ret = Attribute.convert_with_links(attribute)
        LOG.info('[attributes] post returned: %s', ret)
        return ret

    @authorize_wsgi.authorize_wsgi("cyborg:attribute", "delete")
    @expose.expose(None, wtypes.text, status_code=HTTPStatus.NO_CONTENT)
    def delete(self, uuid):
        """Delete one attribute.
            - UUID of a attribute.
        """
        LOG.info('[attributes] delete by uuid: %s.', uuid)
        context = pecan.request.context
        api_obj_attribute = objects.Attribute.get(context, uuid)
        api_obj_attribute.destroy(context)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc


class GenericDriver(object, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def discover(self):
        """Discover a specified accelerator.

        :return: the list of driver device objs
        """
        pass

    @abc.abstractmethod
    def update(self, control_path, image_path):
        """Update the device firmware with specific image.

        :param control_path: the image update control path of device.
        :param image_path: The image path of the firmware binary.

        :return: True if update successfully otherwise False
        """
        pass

    @abc.abstractmethod
    def get_stats(self):
        """Collects device stats.

        It is used to collect information from the device about the device
        capabilities. Such as performance info like temprature, power, volt,
        packet_count info.

        :return: The stats info of the device. The format should follow the
        current Cyborg device-deploy-accelerator model
        """
        pass



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields as object_fields

from cyborg.common import constants

# Import fields from oslo_versionedobjects
EnumField = object_fields.EnumField
IntegerField = object_fields.IntegerField
UUIDField = object_fields.UUIDField
StringField = object_fields.StringField
DateTimeField = object_fields.DateTimeField
BooleanField = object_fields.BooleanField
ObjectField = object_fields.ObjectField
ListOfObjectsField = object_fields.ListOfObjectsField
ListOfStringsField = object_fields.ListOfStringsField
DictOfStringsField = object_fields.DictOfStringsField
IPAddressField = object_fields.IPAddressField
IPNetworkField = object_fields.IPNetworkField
UnspecifiedDefault = object_fields.UnspecifiedDefault
ListOfDictOfNullableStringsField = (
    object_fields.ListOfDictOfNullableStringsField)


class ARQState(object_fields.Enum):
    ALL = constants.ARQ_STATES

    def __init__(self):
        super(ARQState, self).__init__(valid_values=ARQState.ALL)


class ARQStateField(object_fields.BaseEnumField):
    AUTO_TYPE = ARQState()


class DeviceTypeField(object_fields.AutoTypedField):
    AUTO_TYPE = object_fields.Enum(valid_values=constants.DEVICE_TYPE)



# Copyright 2019 Intel, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_versionedobjects import base as object_base

from cyborg.common import constants
from cyborg.db import api as dbapi
from cyborg.objects import base
from cyborg.objects import fields as object_fields

LOG = logging.getLogger(__name__)


@base.CyborgObjectRegistry.register
class AttachHandle(base.CyborgObject, object_base.VersionedObjectDictCompat):
    # Version 1.0: Initial version
    VERSION = '1.0'

    dbapi = dbapi.get_instance()

    fields = {
        'id': object_fields.IntegerField(nullable=False),
        'uuid': object_fields.UUIDField(nullable=False),
        'deployable_id': object_fields.IntegerField(nullable=False),
        'cpid_id': object_fields.IntegerField(nullable=False),
        'attach_type': object_fields.EnumField(
            valid_values=constants.ATTACH_HANDLE_TYPES,
            nullable=False),
        # attach_info should be JSON here.
        'attach_info': object_fields.StringField(nullable=False),
        'in_use': object_fields.BooleanField(nullable=False, default=False)
    }

    def create(self, context):
        """Create a AttachHandle record in the DB."""
        values = self.obj_get_changes()
        db_ah = self.dbapi.attach_handle_create(context, values)
        self._from_db_object(self, db_ah)

    @classmethod
    def get(cls, context, uuid):
        """Find a DB AttachHandle and return an Obj AttachHandle."""
        db_ah = cls.dbapi.attach_handle_get_by_uuid(context, uuid)
        obj_ah = cls._from_db_object(cls(context), db_ah)
        return obj_ah

    @classmethod
    def get_by_id(cls, context, id):
        """Find a DB AttachHandle by ID and return an Obj AttachHandle."""
        db_ah = cls.dbapi.attach_handle_get_by_id(context, id)
        obj_ah = cls._from_db_object(cls(context), db_ah)
        return obj_ah

    @classmethod
    def list(cls, context, filters=None):
        """Return a list of AttachHandle objects."""
        if filters:
            sort_dir = filters.pop('sort_dir', 'desc')
            sort_key = filters.pop('sort_key', 'created_at')
            limit = filters.pop('limit', None)
            marker = filters.pop('marker_obj', None)
            db_ahs = cls.dbapi.attach_handle_get_by_filters(context, filters,
                                                            sort_dir=sort_dir,
                                                            sort_key=sort_key,
                                                            limit=limit,
                                                            marker=marker)
        else:
            db_ahs = cls.dbapi.attach_handle_list(context)
        obj_ah_list = cls._from_db_object_list(db_ahs, context)
        return obj_ah_list

    def save(self, context):
        """Update an AttachHandle record in the DB"""
        updates = self.obj_get_changes()
        db_ahs = self.dbapi.attach_handle_update(context, self.uuid, updates)
        self._from_db_object(self, db_ahs)

    def destroy(self, context):
        """Delete a AttachHandle from the DB."""
        self.dbapi.attach_handle_delete(context, self.uuid)
        self.obj_reset_changes()

    @classmethod
    def get_ah_list_by_deployable_id(cls, context, deployable_id):
        ah_filter = {'deployable_id': deployable_id}
        ah_obj_list = AttachHandle.list(context, ah_filter)
        return ah_obj_list

    @classmethod
    def get_ah_by_depid_attachinfo(cls, context, deployable_id, attach_info):
        ah_filter = {'deployable_id': deployable_id,
                     'attach_info': attach_info}
        ah_obj_list = AttachHandle.list(context, ah_filter)
        if len(ah_obj_list) != 0:
            return ah_obj_list[0]
        else:
            return None

    @classmethod
    def allocate(cls, context, deployable_id):
        db_ah = cls.dbapi.attach_handle_allocate(context, deployable_id)
        obj_ah = cls._from_db_object(cls(context), db_ah)
        return obj_ah

    def deallocate(self, context):
        values = {"in_use": False}
        self.dbapi.attach_handle_update(context, self.uuid, values)



# Copyright 2018 Huawei Technologies Co.,LTD.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from cyborg import objects
from cyborg.objects import fields


def fake_db_deployable(**updates):
    root_uuid = uuidutils.generate_uuid()
    db_deployable = {
        'id': 1,
        'uuid': root_uuid,
        'name': 'dp_name',
        'parent_id': None,
        'root_id': 1,
        'num_accelerators': 4,
        'device_id': 0,
        'driver_name': "fake-driver-name",
        'rp_uuid': None,
        'bitstream_id': None,
        }

    for name, field in objects.Deployable.fields.items():
        if name in db_deployable:
            continue
        if field.nullable:
            db_deployable[name] = None
        elif field.default != fields.UnspecifiedDefault:
            db_deployable[name] = field.default
        else:
            raise Exception('fake_db_deployable needs help with %s' % name)

    if updates:
        db_deployable.update(updates)

    return db_deployable


def fake_deployable_obj(context, obj_dpl_class=None, **updates):
    if obj_dpl_class is None:
        obj_dpl_class = objects.Deployable
    deploy = obj_dpl_class._from_db_object(obj_dpl_class(),
                                           fake_db_deployable(**updates))
    deploy.obj_reset_changes()
    return deploy



__import__('cyborg.objects.driver_objects.driver_device')
__import__('cyborg.objects.driver_objects.driver_attribute')
__import__('cyborg.objects.driver_objects.driver_attach_handle')
__import__('cyborg.objects.driver_objects.driver_deployable')
__import__('cyborg.objects.driver_objects.driver_controlpath_id')



# Copyright 2018 Intel, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
Cyborg Intel FPGA driver implementation.
"""


import glob
import os
import re
import socket

from oslo_serialization import jsonutils

from cyborg.accelerator.common import utils
from cyborg.common import constants
from cyborg.objects.driver_objects import driver_attach_handle
from cyborg.objects.driver_objects import driver_attribute
from cyborg.objects.driver_objects import driver_controlpath_id
from cyborg.objects.driver_objects import driver_deployable
from cyborg.objects.driver_objects import driver_device


PCI_DEVICES_PATH = "/sys/bus/pci/devices"
PCI_DEVICES_PATH_PATTERN = "/sys/bus/pci/devices/*"
# TODO(shaohe) The KNOWN_FPGAS can be configurable.
KNOWN_FPGAS = [("0x8086", "0x09c4")]

SYS_FPGA = "/sys/class/fpga"
DEVICE = "device"
PF = "physfn"
VF = "virtfn*"
BDF_PATTERN = re.compile(
    r"^[a-fA-F\d]{4}:[a-fA-F\d]{2}:[a-fA-F\d]{2}\.[a-fA-F\d]$")

DEVICE_FILE_MAP = {"vendor": "vendor",
                   "device": "product_id"}
DEVICE_FILE_HANDLER = {}
DEVICE_EXPOSED = ["vendor", "device"]

PRODUCT_MAP = {"0x09c4": "PAC_ARRIA10"}

DRIVER_NAME = "intel"


def read_line(filename):
    with open(filename) as f:
        return f.readline().strip()


def is_fpga(p):
    infos = (read_line(os.path.join(p, "vendor")),
             read_line(os.path.join(p, "device")))
    if infos in KNOWN_FPGAS:
        return os.path.realpath(p)


def link_real_path(p):
    return os.path.realpath(
        os.path.join(os.path.dirname(p), os.readlink(p)))


# TODO(s_shogo) This function name should be reconsidered in py3
# env( filter() in py3 returns iterator, not list)
def find_fpgas_by_know_list():
    return filter(
        lambda p: (
            read_line(os.path.join(p, "vendor")),
            read_line(os.path.join(p, "device"))
        ) in KNOWN_FPGAS,
        glob.glob(PCI_DEVICES_PATH_PATTERN))


def get_link_targets(links):
    return map(
        lambda p:
            os.path.realpath(
                os.path.join(os.path.dirname(p), os.readlink(p))),
        links)


def all_fpgas():
    # glob.glob("/sys/class/fpga", "*")
    return set(get_link_targets(find_fpgas_by_know_list())) | set(
        map(lambda p: p.rsplit("/", 2)[0],
            get_link_targets(glob.glob(os.path.join(SYS_FPGA, "*")))))


def all_vf_fpgas():
    return [dev.rsplit("/", 2)[0] for dev in
            glob.glob(os.path.join(SYS_FPGA, "*/device/physfn"))]


def all_pfs_have_vf():
    return list(filter(lambda p: glob.glob(os.path.join(p, "virtfn0")),
                all_fpgas()))


def target_symbolic_map():
    maps = {}
    for f in glob.glob(os.path.join(SYS_FPGA, "*/device")):
        maps[os.path.realpath(f)] = os.path.dirname(f)
    return maps


def bdf_path_map():
    return dict(map(lambda f: (os.path.basename(f), f), all_fpgas()))


def all_vfs_in_pf_fpgas(pf_path):
    return get_link_targets(
        glob.glob(os.path.join(pf_path, "virtfn*")))


def all_pf_fpgas():
    return filter(lambda p: glob.glob(os.path.join(p, "sriov_totalvfs")),
                  all_fpgas())


def is_bdf(bdf):
    return True if BDF_PATTERN.match(bdf) else False


def get_bdf_by_path(path):
    bdf = os.path.basename(path)
    if is_bdf(bdf):
        return bdf
    return os.path.basename(os.readlink(os.path.join(path, "device")))


def get_afu_ids(device_name):
    return map(
        read_line,
        glob.glob(
            os.path.join(
                PCI_DEVICES_PATH_PATTERN, "fpga",
                device_name, "intel-fpga-port.*", "afu_id")
        )
    )


def get_region_ids(device_name):
    return map(
        read_line,
        glob.glob(
            os.path.join(
                SYS_FPGA, device_name,
                "intel-fpga-fme.*", "pr/interface_id")
        )
    )


def get_traits(device_name, product_id, vf=True):
    """Generate traits for devices.
    : param devices_name: name of PF/VF, for example, "intel-fpga-dev.0".
    : param product_id: product id of PF/VF, for example, "0x09c4".
    : param vf: True if device_name is a VF, otherwise False.
    """
    # "CUSTOM_PROGRAMMABLE" not support at present
    traits = []
    if not vf:
        traits.append("CUSTOM_FPGA_INTEL")
        traits.append("CUSTOM_FPGA_INTEL_" + PRODUCT_MAP.get(product_id))
    for i in get_afu_ids(device_name):
        trait = "CUSTOM_FPGA_FUNCTION_ID_INTEL_" + i.upper()
        traits.append(trait)
    for i in get_region_ids(device_name):
        trait = "CUSTOM_FPGA_REGION_INTEL_" + i.upper()
        traits.append(trait)
    return {"traits": traits}


def fpga_device(path):
    infos = {}

    # NOTE "In 3.x, os.path.walk is removed in favor of os.walk."
    for (dirpath, dirnames, filenames) in os.walk(path):
        for filename in filenames:
            if filename in DEVICE_EXPOSED:
                key = DEVICE_FILE_MAP.get(filename) or filename
                if key in DEVICE_FILE_HANDLER and callable(
                        DEVICE_FILE_HANDLER(key)):
                    infos[key] = DEVICE_FILE_HANDLER(key)(
                        os.path.join(dirpath, filename))
                else:
                    infos[key] = read_line(os.path.join(dirpath, filename))
    return infos


def fpga_tree():
    def gen_fpga_infos(path, vf=True):
        bdf = get_bdf_by_path(path)
        names = glob.glob1(os.path.join(path, "fpga"), "*")
        fpga = {"type": constants.DEVICE_FPGA,
                "devices": bdf, "stub": True,
                "name": "_".join((socket.gethostname(), bdf))}
        d_info = fpga_device(path)
        fpga.update(d_info)
        if names:
            name = names[0]
            fpga["stub"] = False
            traits = get_traits(name, fpga["product_id"], vf)
            fpga.update(traits)
        fpga["rc"] = constants.RESOURCES["FPGA"]
        return fpga

    devs = []
    pf_has_vf = all_pfs_have_vf()
    for pf in all_pf_fpgas():
        fpga = gen_fpga_infos(pf, False)
        if pf in pf_has_vf:
            # Currently only one region supported.
            fpga["regions"] = []
            # All VFs here belong to one same region.
            vfs = all_vfs_in_pf_fpgas(pf)
            for vf in vfs:
                vf_fpga = gen_fpga_infos(vf, True)
                fpga["regions"].append(vf_fpga)
        devs.append(_generate_driver_device(fpga, pf in pf_has_vf))
    return devs


def _generate_driver_device(fpga, pf_has_vf):
    driver_device_obj = driver_device.DriverDevice()
    driver_device_obj.vendor = fpga["vendor"]
    driver_device_obj.stub = fpga["stub"]
    driver_device_obj.model = fpga.get('model', "miss_model_info")
    driver_device_obj.vendor_board_info = fpga.get('vendor_board_info',
                                                   "miss_vb_info")
    std_board_info = {'product_id': fpga.get('product_id')}
    driver_device_obj.std_board_info = jsonutils.dumps(std_board_info)
    driver_device_obj.type = fpga["type"]
    driver_device_obj.controlpath_id = _generate_controlpath_id(fpga)
    driver_device_obj.deployable_list = _generate_dep_list(fpga, pf_has_vf)
    return driver_device_obj


def _generate_controlpath_id(fpga):
    driver_cpid = driver_controlpath_id.DriverControlPathID()
    driver_cpid.cpid_type = "PCI"
    driver_cpid.cpid_info = utils.pci_str_to_json(fpga["devices"])
    return driver_cpid


def _generate_dep_list(fpga, pf_has_vf):
    driver_dep = driver_deployable.DriverDeployable()
    driver_dep.attribute_list = _generate_attribute_list(fpga)
    driver_dep.attach_handle_list = []
    # pf without sriov enabled.
    if not pf_has_vf:
        driver_dep.num_accelerators = 1
        driver_dep.attach_handle_list = \
            [_generate_attach_handle(fpga)]
        driver_dep.name = fpga["name"]
        driver_dep.driver_name = DRIVER_NAME
    # pf with sriov enabled, may have several regions and several vfs.
    # For now, there is only one region, this maybe improve in next release.
    else:
        driver_dep.num_accelerators = len(fpga["regions"])
        for vf in fpga["regions"]:
            # Only vfs in regions can be attach, no pf.
            driver_dep.attach_handle_list.append(
                _generate_attach_handle(vf))
            driver_dep.name = vf["name"]
            driver_dep.driver_name = DRIVER_NAME
    return [driver_dep]


def _generate_attach_handle(fpga):
    driver_ah = driver_attach_handle.DriverAttachHandle()
    driver_ah.attach_type = constants.AH_TYPE_PCI
    driver_ah.attach_info = utils.pci_str_to_json(fpga["devices"])
    driver_ah.in_use = False
    return driver_ah


def _generate_attribute_list(fpga):
    attr_list = []
    count_pf_traits = 0
    for k, v in fpga.items():
        if k == "rc":
            driver_attr = driver_attribute.DriverAttribute()
            driver_attr.key, driver_attr.value = k, v
            attr_list.append(driver_attr)
        if k == "traits":
            values = fpga.get(k, [])
            for index_pf, val in enumerate(values):
                driver_attr = driver_attribute.DriverAttribute()
                driver_attr.key = "trait" + str(index_pf)
                driver_attr.value = val
                attr_list.append(driver_attr)
                count_pf_traits = index_pf
    vfs = fpga.get('regions', [])
    for vf in vfs:
        for k, values in vf.items():
            if k == "traits":
                for index_vf, val in enumerate(values):
                    driver_attr = driver_attribute.DriverAttribute()
                    driver_attr.key = "trait" + str(index_vf + count_pf_traits)
                    driver_attr.value = val
                    attr_list.append(driver_attr)
    return attr_list



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Cyborg common internal object model"""

import netaddr
from oslo_log import log as logging
from oslo_utils import versionutils
from oslo_versionedobjects import base as object_base

from cyborg import objects
from cyborg.objects import fields as object_fields


LOG = logging.getLogger(__name__)


class CyborgObjectRegistry(object_base.VersionedObjectRegistry):
    def registration_hook(self, cls, index):
        # NOTE(jroll): blatantly stolen from nova
        # NOTE(danms): This is called when an object is registered,
        # and is responsible for maintaining cyborg.objects.$OBJECT
        # as the highest-versioned implementation of a given object.
        version = versionutils.convert_version_to_tuple(cls.VERSION)
        if not hasattr(objects, cls.obj_name()):
            setattr(objects, cls.obj_name(), cls)
        else:
            cur_version = versionutils.convert_version_to_tuple(
                getattr(objects, cls.obj_name()).VERSION)
            if version >= cur_version:
                setattr(objects, cls.obj_name(), cls)


class CyborgObject(object_base.VersionedObject):
    """Base class and object factory.

    This forms the base of all objects that can be remoted or instantiated
    via RPC. Simply defining a class that inherits from this base class
    will make it remotely instantiatable. Objects should implement the
    necessary "get" classmethod routines as well as "save" object methods
    as appropriate.
    """

    OBJ_SERIAL_NAMESPACE = 'cyborg_object'
    OBJ_PROJECT_NAMESPACE = 'cyborg'

    fields = {
        'created_at': object_fields.DateTimeField(nullable=True),
        'updated_at': object_fields.DateTimeField(nullable=True),
    }

    def as_dict(self):
        """Return the object represented as a dict.
        The returned object is JSON-serialisable.
        """

        def _attr_as_dict(field):
            """Return an attribute as a dict, handling nested objects."""
            attr = getattr(self, field)
            if isinstance(attr, CyborgObject):
                attr = attr.as_dict()
            return attr

        return {k: _attr_as_dict(k)
                for k in self.fields if self.obj_attr_is_set(k)}

    @staticmethod
    def _from_db_object(obj, db_obj):
        """Converts a database entity to a formal object.

        :param obj: An object of the class.
        :param db_obj: A DB model of the object
        :return: The object of the class with the database entity added
        """

        for field in obj.fields:
            obj[field] = db_obj[field]

        obj.obj_reset_changes()
        return obj

    @classmethod
    def _from_db_object_list(cls, db_objs, context):
        """Converts a list of database entities to a list of formal objects."""
        objs = []
        for db_obj in db_objs:
            objs.append(cls._from_db_object(cls(context), db_obj))
        return objs

    def obj_make_compatible(self, primitive, target_version):
        """Make an object representation compatible with a target version.

        This is responsible for taking the primitive representation of
        an object and making it suitable for the given target_version.
        This may mean converting the format of object attributes, removing
        attributes that have been added since the target version, etc.

        :param:primitive: The result of self.obj_to_primitive()
        :param:target_version: The version string requested by the recipient
                               of the object.
        """
        _log_backport(self, target_version)
        super(CyborgObject, self).obj_make_compatible(primitive,
                                                      target_version)


class CyborgObjectSerializer(object_base.VersionedObjectSerializer):
    # Base class to use for object hydration
    OBJ_BASE_CLASS = CyborgObject


CyborgObjectDictCompat = object_base.VersionedObjectDictCompat


class CyborgPersistentObject(object):
    """Mixin class for Persistent objects.

    This adds the fields that we use in common for most persistent objects.
    """
    fields = {
        'created_at': object_fields.DateTimeField(nullable=True),
        'updated_at': object_fields.DateTimeField(nullable=True),
        'deleted_at': object_fields.DateTimeField(nullable=True),
        'deleted': object_fields.BooleanField(default=False),
        }


class ObjectListBase(object_base.ObjectListBase):

    @classmethod
    def _obj_primitive_key(cls, field):
        return 'cyborg_object.%s' % field

    @classmethod
    def _obj_primitive_field(cls, primitive, field,
                             default=object_fields.UnspecifiedDefault):
        key = cls._obj_primitive_key(field)
        if default == object_fields.UnspecifiedDefault:
            return primitive[key]
        else:
            return primitive.get(key, default)


def obj_to_primitive(obj):
    """Recursively turn an object into a python primitive.

    A CyborgObject becomes a dict, and anything that implements ObjectListBase
    becomes a list.
    """
    if isinstance(obj, ObjectListBase):
        return [obj_to_primitive(x) for x in obj]
    elif isinstance(obj, CyborgObject):
        result = {}
        for key in obj.obj_fields:
            if obj.obj_attr_is_set(key) or key in obj.obj_extra_fields:
                result[key] = obj_to_primitive(getattr(obj, key))
        return result
    elif isinstance(obj, netaddr.IPAddress):
        return str(obj)
    elif isinstance(obj, netaddr.IPNetwork):
        return str(obj)
    else:
        return obj


def obj_equal_prims(obj_1, obj_2, ignore=None):
    """Compare two primitives for equivalence ignoring some keys.
    This operation tests the primitives of two objects for equivalence.
    Object primitives may contain a list identifying fields that have been
    changed - this is ignored in the comparison. The ignore parameter lists
    any other keys to be ignored.
    :param:obj1: The first object in the comparison
    :param:obj2: The second object in the comparison
    :param:ignore: A list of fields to ignore
    :returns: True if the primitives are equal ignoring changes
    and specified fields, otherwise False.
    """

    def _strip(prim, keys):
        if isinstance(prim, dict):
            for k in keys:
                prim.pop(k, None)
            for v in prim.values():
                _strip(v, keys)
        if isinstance(prim, list):
            for v in prim:
                _strip(v, keys)
        return prim

    if ignore is not None:
        keys = ['cyborg_object.changes'] + ignore
    else:
        keys = ['cyborg_object.changes']
    prim_1 = _strip(obj_1.obj_to_primitive(), keys)
    prim_2 = _strip(obj_2.obj_to_primitive(), keys)
    return prim_1 == prim_2


class DriverObjectBase(CyborgObject):
    @staticmethod
    def _from_db_object(obj, db_obj):
        fields = obj.fields
        fields.pop("updated_at")
        fields.pop("created_at")
        for field in fields:
            obj[field] = db_obj[field]

        obj.obj_reset_changes()
        return obj


def _log_backport(ovo, target_version):
    """Log backported versioned objects."""
    if target_version and target_version != ovo.VERSION:
        LOG.debug('Backporting %(obj_name)s from version %(src_vers)s '
                  'to version %(dst_vers)s',
                  {'obj_name': ovo.obj_name(),
                   'src_vers': ovo.VERSION,
                   'dst_vers': target_version})



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Base classes for API tests."""

from oslo_config import cfg
from oslo_context import context
import pecan
import pecan.testing

from cyborg.tests.unit.db import base

cfg.CONF.import_group('keystone_authtoken', 'keystonemiddleware.auth_token')


class BaseApiTest(base.DbTestCase):
    """Pecan controller functional testing class.

    Used for functional tests of Pecan controllers where you need to
    test your literal application and its integration with the
    framework.
    """

    PATH_PREFIX = ''

    def setUp(self):
        super(BaseApiTest, self).setUp()
        cfg.CONF.set_override("auth_version", "v3",
                              group='keystone_authtoken')
        cfg.CONF.set_override("admin_user", "admin",
                              group='keystone_authtoken')
        self.app = self._make_app()

        def reset_pecan():
            pecan.set_config({}, overwrite=True)

        self.addCleanup(reset_pecan)

    def flags(self, **kw):
        """Override flag variables for a test."""
        group = kw.pop('group', None)
        for k, v in kw.items():
            cfg.CONF.set_override(k, v, group)

    def _make_app(self):
        # Determine where we are so we can set up paths in the config
        root_dir = self.get_path()

        self.app_config = {
            'app': {
                'root': 'cyborg.api.controllers.root.RootController',
                'modules': ['cyborg.api'],
                'static_root': '%s/public' % root_dir,
                'template_path': '%s/api/templates' % root_dir,
                'acl_public_routes': ['/', '/v1/.*'],
            },
        }
        return pecan.testing.load_test_app(self.app_config)

    def _request_json(self, path, params, expect_errors=False, headers=None,
                      method="post", extra_environ=None, status=None):
        """Sends simulated HTTP request to Pecan test app.

        :param path: url path of target service
        :param params: content for wsgi.input of request
        :param expect_errors: Boolean value; whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param method: Request method type. Appropriate method function call
                       should be used rather than passing attribute in.
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param status: expected status code of response
        """
        response = getattr(self.app, "%s_json" % method)(
            str(path),
            params=params,
            headers=headers,
            status=status,
            extra_environ=extra_environ,
            expect_errors=expect_errors
        )
        return response

    def post_json(self, path, params, expect_errors=False, headers=None,
                  extra_environ=None, status=None):
        """Sends simulated HTTP POST request to Pecan test app.

        :param path: url path of target service
        :param params: content for wsgi.input of request
        :param expect_errors: Boolean value; whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param status: expected status code of response
        """
        full_path = self.PATH_PREFIX + path
        return self._request_json(path=full_path, params=params,
                                  expect_errors=expect_errors,
                                  headers=headers, extra_environ=extra_environ,
                                  status=status, method="post")

    def gen_context(self, value, **kwargs):
        ct = context.RequestContext.from_dict(value, **kwargs)
        return ct

    def gen_headers(self, context, **kw):
        """Generate a header for a simulated HTTP request to Pecan test app.

        :param context: context that store the client user information.
        :param kw: key word aguments, used to overwrite the context attribute.

        note: "is_public_api" is not in headers, it should be in environ
        variables to send along with the request. We can pass it by
        extra_environ when we call delete, get_json or other method request.
        """
        ct = context.to_dict()
        ct.update(kw)
        if ct.get("is_admin"):
            role = "admin"
        else:
            role = "user"
        headers = {
            'X-User-Name': ct.get("user_name") or "user",
            'X-User-Id':
                ct.get("user_id") or "1d6d686bc2c949ddb685ffb4682e0047",
            'X-Project-Name': ct.get("project_name") or "no_project_name",
            'X-Project-Id':
                ct.get("project_id") or "86f64f561b6d4f479655384572727f70",
            'X-User-Domain-Id':
                ct.get("domain_id") or "bd5eeb7d0fb046daaf694b36f4df5518",
            'X-User-Domain-Name': ct.get("domain_name") or "no_domain",
            'X-Auth-Token':
                ct.get("auth_token") or "b9764005b8c145bf972634fb16a826e8",
            'X-Roles': ct.get("roles") or role,
        }
        if ct.get('system_scope') == 'all':
            headers.update({'Openstack-System-Scope': 'all'})
        return headers

    def get_json(self, path, expect_errors=False, headers=None,
                 extra_environ=None, q=None, return_json=True, **params):
        """Sends simulated HTTP GET request to Pecan test app.

        :param path: url path of target service
        :param expect_errors: Boolean value; whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param q: list of queries consisting of: field, value, op, and type
                  keys
        :param path_prefix: prefix of the url path
        :param params: content for wsgi.input of request
        """
        full_path = self.PATH_PREFIX + path
        q = q or []
        query_params = {
            'q.field': [],
            'q.value': [],
            'q.op': [],
            }
        for query in q:
            for name in ['field', 'op', 'value']:
                query_params['q.%s' % name].append(query.get(name, ''))
        all_params = {}
        all_params.update(params)
        if q:
            all_params.update(query_params)
        response = self.app.get(full_path,
                                params=all_params,
                                headers=headers,
                                extra_environ=extra_environ,
                                expect_errors=expect_errors)
        if return_json and not expect_errors:
            response = response.json
        return response

    def patch_json(self, path, params, expect_errors=False, headers=None,
                   extra_environ=None, status=None):
        """Sends simulated HTTP PATCH request to Pecan test app.

        :param path: url path of target service
        :param params: content for wsgi.input of request
        :param expect_errors: Boolean value; whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param status: expected status code of response
        """
        full_path = self.PATH_PREFIX + path
        return self._request_json(path=full_path, params=params,
                                  expect_errors=expect_errors,
                                  headers=headers, extra_environ=extra_environ,
                                  status=status, method="patch")

    def delete(self, path, expect_errors=False, headers=None,
               extra_environ=None, status=None):
        """Sends simulated HTTP DELETE request to Pecan test app.

        :param path: url path of target service
        :param expect_errors: Boolean value; whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param status: expected status code of response
        """
        full_path = self.PATH_PREFIX + path
        response = self.app.delete(full_path,
                                   headers=headers,
                                   status=status,
                                   extra_environ=extra_environ,
                                   expect_errors=expect_errors)
        return response



# Copyright 2019 Intel Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_serialization import jsonutils
from oslo_utils import timeutils
from oslo_utils import uuidutils

from cyborg.common import constants
from cyborg.common import exception
from cyborg.common import utils
from cyborg.conf import CONF
from cyborg import objects
from cyborg.tests.unit.db import base
from cyborg.tests.unit import fake_deployable
from cyborg.tests.unit import fake_extarq


class TestExtARQJobMixin(base.DbTestCase):

    def setUp(self):
        super(TestExtARQJobMixin, self).setUp()
        self.fake_db_extarqs = fake_extarq.get_fake_db_extarqs()
        self.fake_obj_extarqs = fake_extarq.get_fake_extarq_objs()
        self.fake_obj_fpga_extarqs = fake_extarq.get_fake_fpga_extarq_objs()
        self.deployable_uuids = ['0acbf8d6-e02a-4394-aae3-57557d209498']
        self.classes = ["gpu", "no_program", "bitstream_program",
                        "function_program", "bad_program"]
        self.class_objects = dict(
            zip(self.classes, self.fake_obj_extarqs))
        self.class_dbs = dict(
            zip(self.classes, self.fake_db_extarqs))
        self.fpga_classes = ["no_program", "bitstream_program",
                             "function_program", "bad_program"]
        self.fpga_class_objects = dict(
            zip(self.fpga_classes, self.fake_obj_fpga_extarqs))
        self.bitstream_id = self.class_objects["bitstream_program"][
            "device_profile_group"][constants.ACCEL_BITSTREAM_ID]
        self.function_id = self.class_objects["function_program"][
            "device_profile_group"][constants.ACCEL_FUNCTION_ID]

    def test_get_resources_from_device_profile_group(self):
        expect = [("GPU")] + [("FPGA")] * 4
        actual = [v.get_resources_from_device_profile_group()
                  for v in self.class_objects.values()]
        self.assertEqual(expect, actual)

    def test_get_suitable_ext_arq(self):
        expect_type = [objects.ExtARQ] + [objects.FPGAExtARQ] * 4
        uuid = uuidutils.generate_uuid()
        groups = [v['device_profile_group']
                  for v in self.fake_db_extarqs]
        dp_db = {
            "id": self.fake_db_extarqs[0]['device_profile_id'],
            "uuid": uuid,
            "name": self.fake_db_extarqs[0]['device_profile_name'],
            "description": "fake-device_profile_desc",
            "profile_json": jsonutils.dumps({"groups": groups}),
            "created_at": timeutils.utcnow().isoformat(),
            "updated_at": timeutils.utcnow().isoformat(),
        }
        for i, v in enumerate(self.classes):
            with mock.patch.object(self.dbapi, 'extarq_get') \
                as mock_extarq_get, \
                mock.patch.object(self.dbapi,
                                  "device_profile_get_by_id") \
                as mock_dp_get, \
                mock.patch.object(self.dbapi,
                                  "attach_handle_get_by_id") \
                as mock_ah_get:
                mock_ah_get.return_value = None
                mock_dp_get.return_value = dp_db
                self.fake_db_extarqs[i].update({
                    "state": constants.ARQ_BIND_STARTED,
                    "substate": constants.ARQ_BIND_STARTED,
                    "attach_handle_id": None,
                    "created_at": timeutils.utcnow().isoformat(),
                    "updated_at": timeutils.utcnow().isoformat()
                })
                self.fake_db_extarqs[i]["deployable_id"] = None
                mock_extarq_get.return_value = self.fake_db_extarqs[i]
                obj = self.class_objects[v]
                uuid = obj.arq.uuid
                typ = obj.get_suitable_ext_arq(self.context, uuid)
                self.assertIsInstance(typ, expect_type[i])

    def test_start_bind_job_raise(self):
        obj_extarq = self.class_objects["bitstream_program"]
        instance_uuid = obj_extarq.arq.instance_uuid
        uuid = obj_extarq.arq.uuid
        valid_fields = {
            uuid: {'hostname': obj_extarq.arq.hostname,
                   'device_rp_uuid': obj_extarq.arq.device_rp_uuid,
                   'instance_uuid': instance_uuid}
            }

        self.assertRaises(
            exception.ARQBadState, obj_extarq.start_bind_job,
            self.context, valid_fields)

    @mock.patch('cyborg.objects.extarq.ext_arq_job.ExtARQJobMixin._bind_job')
    @mock.patch('cyborg.objects.deployable.Deployable.get_by_device_rp_uuid')
    @mock.patch('cyborg.objects.ExtARQ.update_check_state')
    def test_start_bind_job(self, mock_state, mock_get_dep, mock_job):
        obj_extarq = self.class_objects["bitstream_program"]
        obj_extarq.arq.state = constants.ARQ_UNBOUND
        instance_uuid = obj_extarq.arq.instance_uuid
        uuid = obj_extarq.arq.uuid
        dep_uuid = self.deployable_uuids[0]
        fake_dep = fake_deployable.fake_deployable_obj(self.context,
                                                       uuid=dep_uuid)
        mock_get_dep.return_value = fake_dep

        valid_fields = {
            uuid: {'hostname': obj_extarq.arq.hostname,
                   'device_rp_uuid': obj_extarq.arq.device_rp_uuid,
                   'instance_uuid': instance_uuid}
            }
        obj_extarq.start_bind_job(self.context, valid_fields)
        mock_state.assert_called_once_with(
            self.context, constants.ARQ_BIND_STARTED)
        mock_job.assert_called_once_with(self.context, fake_dep)

    @mock.patch('cyborg.objects.FPGAExtARQ.bind')
    def test_gpu_arq_start_bind_job(self, mock_aysnc_bind):
        # Test GPU ARQ does not need async bind
        obj_extarq = self.class_objects["gpu"]
        obj_extarq.arq.state = constants.ARQ_UNBOUND
        dep_uuid = self.deployable_uuids[0]
        fake_dep = fake_deployable.fake_deployable_obj(self.context,
                                                       uuid=dep_uuid)
        is_job = getattr(obj_extarq.bind, "is_job", False)
        with mock.patch.object(obj_extarq, 'bind') as mock_bind:
            mock_bind.is_job = is_job
            obj_extarq._bind_job(self.context, fake_dep)
            mock_bind.assert_called_once_with(self.context, fake_dep)
            mock_aysnc_bind.assert_not_called()

    @mock.patch('cyborg.objects.ExtARQ.bind')
    def test_fpga_sync_start_bind_job(self, mock_bind):
        # Test FPGA with no program does not need async bind
        obj_extarq = self.fpga_class_objects["no_program"]
        obj_extarq.arq.state = constants.ARQ_UNBOUND
        dep_uuid = self.deployable_uuids[0]
        fake_dep = fake_deployable.fake_deployable_obj(self.context,
                                                       uuid=dep_uuid)
        need_bind = getattr(obj_extarq.bind, "is_job", False)
        with mock.patch.object(obj_extarq, 'bind') as mock_aysnc_bind:
            mock_aysnc_bind.is_job = need_bind
            obj_extarq._bind_job(self.context, fake_dep)
            mock_aysnc_bind.assert_called_once_with(self.context, fake_dep)
            mock_bind.assert_not_called()

    @mock.patch('cyborg.objects.ExtARQ.bind')
    @mock.patch('cyborg.common.utils.ThreadWorks.spawn')
    def test_fpga_async_start_bind_job(self, mock_spawn, mock_bind):
        # Test FPGA with bitstream program need async bind
        obj_extarq = self.fpga_class_objects["bitstream_program"]
        obj_extarq.arq.state = constants.ARQ_UNBOUND
        dep_uuid = self.deployable_uuids[0]
        fake_dep = fake_deployable.fake_deployable_obj(self.context,
                                                       uuid=dep_uuid)
        need_bind = getattr(obj_extarq.bind, "is_job", False)
        with mock.patch.object(obj_extarq, 'bind') as mock_aysnc_bind:
            mock_aysnc_bind.is_job = need_bind
            obj_extarq._bind_job(self.context, fake_dep)
            mock_spawn.assert_called_once_with(
                mock_aysnc_bind, self.context, fake_dep)
            mock_bind.assert_not_called()

    @mock.patch('cyborg.common.utils.ThreadWorks.get_workers_result')
    @mock.patch('cyborg.common.utils.ThreadWorks.spawn_master')
    def test_master_with_async_jobs(self, mock_spawn, mock_get_result):
        # There are async job, so will start to monitor the job status
        def job(context, deployable):
            pass

        works = utils.ThreadWorks()
        dep_uuid = self.deployable_uuids[0]
        fake_dep = fake_deployable.fake_deployable_obj(self.context,
                                                       uuid=dep_uuid)
        arq_job_binds = {
            self.class_objects["bitstream_program"]:
            works.spawn(job, self.context, fake_dep),
            self.class_objects["function_program"]:
            works.spawn(job, self.context, fake_dep)
        }
        arq_binds = {
            self.class_objects["gpu"]: None,
            self.class_objects["no_program"]: None,
        }
        arq_binds.update(arq_job_binds)
        objects.ext_arq.ExtARQJobMixin.master(self.context, arq_binds)
        mock_get_result.return_value = "Jobs_Generator"
        mock_get_result.assert_called_once()
        mock_spawn.assert_called_once()

    @mock.patch('cyborg.objects.ext_arq.ExtARQJobMixin.check_bindings_result')
    def test_master_with_instant(self, mock_result):
        # There are no async job, so will not start to monitor the job status
        arq_binds = {
            self.class_objects["gpu"]: None,
            self.class_objects["no_program"]: None,
        }
        objects.ext_arq.ExtARQ.master(self.context, arq_binds)
        mock_result.assert_called_once_with(self.context, arq_binds.keys())

    @mock.patch('cyborg.objects.ExtARQ.bind_notify')
    @mock.patch('cyborg.objects.ExtARQ.list')
    def test_check_bindings_result_with_arq_deleted(
        self, mock_list, mock_notify):
        bind_status = [
            (e.arq.uuid, constants.ARQ_BIND_STATES_STATUS_MAP[e.arq.state])
            for e in self.fake_obj_extarqs]
        extarqs = self.fake_obj_extarqs
        mock_list.return_value = self.fake_obj_fpga_extarqs
        instance_uuid = extarqs[0].arq.instance_uuid
        objects.ExtARQ.check_bindings_result(
            self.context, extarqs)
        mock_notify.assert_called_once_with(instance_uuid, bind_status)

    @mock.patch('cyborg.objects.ExtARQ.bind_notify')
    @mock.patch('cyborg.objects.ExtARQ.list')
    def test_check_bindings_result_with_non_bound_arq(
        self, mock_list, mock_notify):
        bind_status = [
            (e.arq.uuid, constants.ARQ_BIND_STATES_STATUS_MAP[e.arq.state])
            for e in self.fake_obj_extarqs]
        extarqs = self.fake_obj_extarqs
        mock_list.return_value = extarqs
        instance_uuid = extarqs[0].arq.instance_uuid
        states = [constants.ARQ_BIND_STARTED, constants.ARQ_BIND_STARTED,
                  constants.ARQ_UNBOUND, constants.ARQ_BOUND,
                  constants.ARQ_BIND_FAILED, constants.ARQ_DELETING]
        # Completed
        objects.ExtARQ.check_bindings_result(self.context, extarqs)
        mock_notify.assert_called_once_with(instance_uuid, bind_status)
        # Exception
        for s in states[:2]:
            extarqs[0].arq.state = s
            self.assertRaises(
                exception.ARQBadState, objects.ExtARQ.check_bindings_result,
                self.context, extarqs)
        # Failed
        for s in states[4:]:
            extarqs[0].arq.state = s
            bind_status[0] = (bind_status[0][0], "failed")
            objects.ExtARQ.check_bindings_result(self.context, extarqs)
            mock_notify.assert_called_with(instance_uuid, bind_status)
        # Completed
        bind_status[0] = (bind_status[0][0], "completed")
        extarqs[0].arq.state = states[3]
        objects.ExtARQ.check_bindings_result(self.context, extarqs)
        mock_notify.assert_called_with(instance_uuid, bind_status)

    @mock.patch('cyborg.objects.ExtARQ.bind_notify')
    @mock.patch('cyborg.objects.ExtARQ.list')
    def test_check_bindings_result_with_arq_bound(
        self, mock_list, mock_notify):
        bind_status = [
            (e.arq.uuid, constants.ARQ_BIND_STATES_STATUS_MAP[e.arq.state])
            for e in self.fake_obj_extarqs]
        extarqs = self.fake_obj_extarqs
        mock_list.return_value = extarqs
        instance_uuid = extarqs[0].arq.instance_uuid
        objects.ExtARQ.check_bindings_result(
            self.context, extarqs)
        mock_notify.assert_called_once_with(instance_uuid, bind_status)

    @mock.patch('cyborg.objects.ext_arq.ExtARQJobMixin.check_bindings_result')
    def test_job_monitor_with_job_exception(self, mock_result):
        works = utils.ThreadWorks()
        err_job = works.spawn(lambda x: x / 0, 1)
        good_job = works.spawn(lambda x: x, 1)
        works_generator = works.get_workers_result(
            [err_job, good_job], timeout=CONF.bind_timeout)

        extarqs = [self.class_objects["bitstream_program"],
                   self.class_objects["function_program"]]

        objects.ext_arq.ExtARQ.job_monitor(
            self.context, works_generator, extarqs)
        mock_result.assert_called_once_with(self.context, extarqs)

    @mock.patch('cyborg.objects.ext_arq.ExtARQJobMixin.check_bindings_result')
    def test_job_monitor_with_job_successful(self, mock_result):
        works = utils.ThreadWorks()
        job1 = works.spawn(lambda x: x, 1)
        job2 = works.spawn(lambda x: x, 2)
        works_generator = works.get_workers_result(
            [job1, job2], timeout=CONF.bind_timeout)

        extarqs = [self.class_objects["bitstream_program"],
                   self.class_objects["function_program"]]

        objects.ext_arq.ExtARQ.job_monitor(
            self.context, works_generator, extarqs)
        mock_result.assert_called_once_with(self.context, extarqs)

    @mock.patch('cyborg.objects.ext_arq.ExtARQJobMixin.check_bindings_result')
    def test_job_monitor_without_jobs(self, mock_result):
        works = utils.ThreadWorks()
        works_generator = works.get_workers_result(
            [], timeout=CONF.bind_timeout)

        extarqs = [self.class_objects["bitstream_program"],
                   self.class_objects["function_program"]]

        objects.ext_arq.ExtARQ.job_monitor(
            self.context, works_generator, extarqs)
        mock_result.assert_called_once_with(self.context, extarqs)

    @mock.patch('cyborg.objects.ext_arq.ExtARQJobMixin.check_bindings_result')
    def test_job_monitor_with_empty_arq(self, mock_result):
        works = utils.ThreadWorks()
        good_job = works.spawn(lambda x: x, 1)
        works_generator = works.get_workers_result(
            [good_job], timeout=CONF.bind_timeout)
        extarqs = []

        objects.ext_arq.ExtARQ.job_monitor(
            self.context, works_generator, extarqs)
        mock_result.assert_not_called()

    @mock.patch('cyborg.common.nova_client.NovaAPI.notify_binding')
    @mock.patch('cyborg.common.nova_client.NovaAPI')
    def test_bind_notify(self, mock_api, mock_notify):
        mock_api.return_value = type(
            "NovaAPI", (object,), {"notify_binding": mock_notify})
        objects.ext_arq.ExtARQJobMixin.bind_notify(
            '5922a70f-1e06-4cfd-88dd-a332120d7144',
            [('a097fefa-da62-4630-8e8b-424c0e3426dc', 'completed')])
        mock_api.assert_called_once_with()
        mock_notify.assert_called_once_with(
            '5922a70f-1e06-4cfd-88dd-a332120d7144',
            [('a097fefa-da62-4630-8e8b-424c0e3426dc', 'completed')])

    @mock.patch('cyborg.objects.ExtARQ.unbind')
    @mock.patch('cyborg.objects.ext_arq.ExtARQJobMixin.get_suitable_ext_arq')
    def test_apply_patch_with_op_remove(self, mock_get, mock_unbind):
        patch_list = {}
        host_binding = {'path': '/hostname', 'op': 'remove',
                        'value': 'myhost'}
        inst_binding = {'path': '/instance_uuid', 'op': 'remove',
                        'value': '5922a70f-1e06-4cfd-88dd-a332120d7144'}
        device_rp_uuid = 'fb16c293-5739-4c84-8590-926f9ab16669'
        arp_uuid = 'a097fefa-da62-4630-8e8b-424c0e3426dc'
        patch_list[arp_uuid] = [host_binding, inst_binding, device_rp_uuid]
        mock_get.return_value = self.fake_obj_extarqs[0]
        valid_fields = {
            arp_uuid: {'hostname': 'myhost',
                       'device_rp_uuid': device_rp_uuid,
                       'instance_uuid': '5922a70f-1e06-4cfd-88dd-a332120d7144'}
        }
        objects.extarq.ext_arq_job.ExtARQJobMixin.apply_patch(self.context,
                                                              patch_list,
                                                              valid_fields)
        mock_unbind.assert_called_with(self.context)



"""add_qat_type

Revision ID: 7e6f1f107f2b
Revises: c1b5abada09c
Create Date: 2019-07-17 04:21:52.055863

"""

# revision identifiers, used by Alembic.
revision = '7e6f1f107f2b'
down_revision = '60d8ac91fd20'

from alembic import op
import sqlalchemy as sa


def upgrade():
    new_device_type = sa.Enum('GPU', 'FPGA', 'AICHIP', 'QAT',
                              name='device_type')
    op.alter_column('devices', 'type',
                    existing_type=new_device_type,
                    nullable=False)



"""add_aichip_type

Revision ID: 589ff20545b7
Revises: ede4e3f1a232
Create Date: 2019-05-22 06:01:08.292535

"""

# revision identifiers, used by Alembic.
revision = '589ff20545b7'
down_revision = 'ede4e3f1a232'

from alembic import op
import sqlalchemy as sa


def upgrade():
    new_device_type = sa.Enum('GPU', 'FPGA', 'AICHIP',
                              name='device_type')
    op.alter_column('devices', 'type',
                    existing_type=new_device_type,
                    nullable=False)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

from cyborg.accelerator.drivers.qat.intel import sysinfo

from cyborg.accelerator.drivers.qat.intel.driver import IntelQATDriver
from cyborg.tests import base
from cyborg.tests.unit.accelerator.drivers.qat.intel import prepare_test_data

import fixtures


class TestIntelQATDriver(base.TestCase):
    def setUp(self):
        super(TestIntelQATDriver, self).setUp()
        self.pcipath = sysinfo.PCI_DEVICES_PATH
        tmp_sys_dir = self.useFixture(fixtures.TempDir())
        prepare_test_data.create_fake_sysfs(tmp_sys_dir.path)
        tmp_path = tmp_sys_dir.path
        sysinfo.PCI_DEVICES_PATH = os.path.join(
            tmp_path, sysinfo.PCI_DEVICES_PATH.split("/", 1)[-1])

    def tearDown(self):
        super(TestIntelQATDriver, self).tearDown()
        sysinfo.PCI_DEVICES_PATH = self.pcipath

    def test_discover(self):
        attach_handle_list = [
            [
                {'attach_type': 'PCI',
                 'attach_info': '{"bus": "05", '
                                '"device": "01", '
                                '"domain": "0000", '
                                '"function": "0"}',
                 'in_use': False}
            ],
            [
                {'attach_type': 'PCI',
                 'attach_info': '{"bus": "06", '
                                '"device": "00", '
                                '"domain": "0000", '
                                '"function": "0"}',
                 'in_use': False}
            ]
        ]
        expected = [{'vendor': '0x8086',
                     'type': 'QAT',
                     'deployable_list':
                         [
                             {'num_accelerators': 1,
                              'name': '0000:05:01.0',
                              'attach_handle_list': attach_handle_list[0]
                              },
                         ],
                     'controlpath_id':
                         {
                             'cpid_info': '{"bus": "05", '
                                          '"device": "00", '
                                          '"domain": "0000", '
                                          '"function": "0"}',
                             'cpid_type': 'PCI'}
                     },
                    {'vendor': '0x8086',
                     'type': 'QAT',
                     'deployable_list':
                         [
                             {'num_accelerators': 1,
                              'name': '0000:06:00.0',
                              'attach_handle_list': attach_handle_list[1]
                              },
                         ],
                     'controlpath_id':
                         {
                             'cpid_info': '{"bus": "06", '
                                          '"device": "00", '
                                          '"domain": "0000", '
                                          '"function": "0"}',
                             'cpid_type': 'PCI'}
                     }
                    ]
        intel = IntelQATDriver()
        qats = intel.discover()
        list.sort(qats, key=lambda x: x._obj_deployable_list[0].name)
        self.assertEqual(2, len(qats))
        for i in range(len(qats)):
            qat_dict = qats[i].as_dict()
            qat_dep_list = qat_dict['deployable_list']
            qat_attach_handle_list = \
                qat_dep_list[0].as_dict()['attach_handle_list']
            self.assertEqual(expected[i]['vendor'], qat_dict['vendor'])
            self.assertEqual(expected[i]['controlpath_id'],
                             qat_dict['controlpath_id'])
            self.assertEqual(expected[i]['deployable_list'][0]
                             ['num_accelerators'],
                             qat_dep_list[0].as_dict()['num_accelerators'])
            self.assertEqual(1, len(qat_attach_handle_list))
            self.assertEqual(attach_handle_list[i][0],
                             qat_attach_handle_list[0].as_dict())



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The Cyborg Conductor Service."""

import sys

from oslo_config import cfg
from oslo_service import service

from cyborg.common import constants
from cyborg.common import service as cyborg_service


CONF = cfg.CONF


def main():
    # Parse config file and command line options, then start logging
    cyborg_service.prepare_service(sys.argv)

    mgr = cyborg_service.RPCService('cyborg.conductor.manager',
                                    'ConductorManager',
                                    constants.CONDUCTOR_TOPIC)

    launcher = service.launch(CONF, mgr, restart_method='mutate')
    launcher.wait()



# Modifications Copyright (C) 2021 ZTE Corporation
# Copyright 2018 Beijing Lenovo Software Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
Cyborg NVIDIA GPU driver implementation.
"""
from oslo_log import log as logging
from oslo_serialization import jsonutils

import collections
import os

import cyborg.conf

from cyborg.accelerator.common import utils
from cyborg.accelerator.drivers.gpu import utils as gpu_utils
from cyborg.common import constants
from cyborg.common import exception
from cyborg.conf import CONF
from cyborg.objects.driver_objects import driver_attach_handle
from cyborg.objects.driver_objects import driver_attribute
from cyborg.objects.driver_objects import driver_controlpath_id
from cyborg.objects.driver_objects import driver_deployable
from cyborg.objects.driver_objects import driver_device

LOG = logging.getLogger(__name__)


def _get_traits(vendor_id, product_id, vgpu_type_name=None):
    """Generate traits for GPUs.
    : param vendor_id: vendor_id of PGPU/VGPU, eg."10de"
    : param product_id: product_id of PGPU/VGPU, eg."1eb8".
    : param vgpu_type_name: vgpu type name, eg."T4_1B".
    Example VGPU traits:
    {traits:["OWNER_CYBORG", "CUSTOM_NVIDIA_1EB8_T4_2B"]}
    Example PGPU traits:
    {traits:["OWNER_CYBORG", "CUSTOM_NVIDIA_1EB8"]}
    """
    traits = ["OWNER_CYBORG"]
    # PGPU trait
    gpu_trait = "_".join(
        ('CUSTOM', gpu_utils.VENDOR_MAPS.get(vendor_id, "").upper(),
         product_id.upper()))
    # VGPU trait
    if vgpu_type_name:
        gpu_trait = "_".join((gpu_trait, vgpu_type_name.upper()))
    traits.append(gpu_trait)
    return {"traits": traits}


def _generate_attribute_list(gpu):
    attr_list = []
    index = 0
    for k, v in gpu.items():
        if k == "rc":
            driver_attr = driver_attribute.DriverAttribute()
            driver_attr.key, driver_attr.value = k, v
            attr_list.append(driver_attr)
        if k == "traits":
            values = gpu.get(k, [])
            for val in values:
                driver_attr = driver_attribute.DriverAttribute(
                    key="trait" + str(index), value=val)
                index = index + 1
                attr_list.append(driver_attr)
    return attr_list


def _generate_attach_handle(gpu, num=None):
    driver_ah = driver_attach_handle.DriverAttachHandle()
    driver_ah.in_use = False
    if gpu["rc"] == "PGPU":
        driver_ah.attach_type = constants.AH_TYPE_PCI
        driver_ah.attach_info = utils.pci_str_to_json(gpu["devices"])
    else:
        vgpu_mark = gpu["vGPU_type"] + '_' + str(num)
        driver_ah.attach_type = constants.AH_TYPE_MDEV
        driver_ah.attach_info = utils.mdev_str_to_json(
            gpu["devices"], gpu["vGPU_type"], vgpu_mark)
    return driver_ah


def _generate_dep_list(gpu):
    driver_dep = driver_deployable.DriverDeployable()
    driver_dep.attribute_list = _generate_attribute_list(gpu)
    driver_dep.attach_handle_list = []
    # NOTE(wangzhh): The name of deployable should be unique, its format is
    # under disscussion, may looks like
    # <ComputeNodeName>_<NumaNodeName>_<CyborgName>_<NumInHost>
    # NOTE(yumeng) Since Wallaby release, the deplpyable_name is named as
    # <Compute_hostname>_<Device_address>
    driver_dep.name = gpu.get('hostname', '') + '_' + gpu["devices"]
    driver_dep.driver_name = \
        gpu_utils.VENDOR_MAPS.get(gpu["vendor_id"], '').upper()
    # if is pGPU, num_accelerators = 1
    if gpu["rc"] == "PGPU":
        driver_dep.num_accelerators = 1
        driver_dep.attach_handle_list = \
            [_generate_attach_handle(gpu)]
    else:
        # if is vGPU, num_accelerators is the total vGPU capability of
        # the asked vGPU type
        vGPU_path = os.path.expandvars(
            '/sys/bus/pci/devices/{0}/mdev_supported_types/{1}/'
            .format(gpu["devices"], gpu["vGPU_type"]))
        num_available = 0
        with open(vGPU_path + 'available_instances', 'r') as f:
            num_available = int(f.read().strip())
        num_created = len(os.listdir(vGPU_path + 'devices'))
        driver_dep.num_accelerators = num_available + num_created
        # example: 1 pGPU has 16 vGPUs is represented as
        # 16 attach_handles, 1 deployable, 1 resource_provider
        # NOTE(yumeng): cyborg use attach_handle_uuid
        # to create each vGPU without the need to generate a new uuid
        # example: echo "attach_handle_uuid" > nvidia-223/create
        for num in range(driver_dep.num_accelerators):
            driver_dep.attach_handle_list.append(
                _generate_attach_handle(gpu, num))
    return [driver_dep]


def _generate_controlpath_id(gpu):
    driver_cpid = driver_controlpath_id.DriverControlPathID()
    driver_cpid.cpid_type = "PCI"
    driver_cpid.cpid_info = utils.pci_str_to_json(gpu["devices"])
    return driver_cpid


def _generate_driver_device(gpu):
    driver_device_obj = driver_device.DriverDevice()
    driver_device_obj.vendor = gpu['vendor_id']
    driver_device_obj.model = gpu.get('model', 'miss model info')
    std_board_info = {'product_id': gpu.get('product_id'),
                      'controller': gpu.get('controller'), }
    vendor_board_info = {'vendor_info': gpu.get('vendor_info',
                         'gpu_vb_info')}
    driver_device_obj.std_board_info = jsonutils.dumps(std_board_info)
    driver_device_obj.vendor_board_info = jsonutils.dumps(
        vendor_board_info)
    driver_device_obj.type = constants.DEVICE_GPU
    driver_device_obj.stub = gpu.get('stub', False)
    driver_device_obj.controlpath_id = _generate_controlpath_id(gpu)
    driver_device_obj.deployable_list = _generate_dep_list(gpu)
    return driver_device_obj


def _get_supported_vgpu_types():
    """Gets supported vgpu_types from cyborg.conf.

    Retrieves supported vgpu_types set by the operator and generates a
    record of vgpu_type and pgpu in the dict constant: pgpu_type_mapping.

    Returns:
        A list of all vgpu_types set in CONF.gpu_devices.enabled_vgpu_types.

    Raises:
        InvalidGPUConfig: An error occurred if same PCI appear twice
        or PCI address is not valid.
    """
    pgpu_type_mapping = collections.defaultdict(str)
    pgpu_type_mapping.clear()
    if not CONF.gpu_devices.enabled_vgpu_types:
        return [], pgpu_type_mapping

    for vgpu_type in CONF.gpu_devices.enabled_vgpu_types:
        group = getattr(CONF, 'vgpu_%s' % vgpu_type, None)
        if group is None or not group.device_addresses:
            # Device addresses must be configured explictly now for every
            # enabled vgpu type. Will improve after the disable and enable
            # devices interfaces implemented.
            raise exception.InvalidvGPUConfig(
                reason="Missing device addresses config for vgpu type %s"
                % vgpu_type
            )
        for device_address in group.device_addresses:
            if device_address in pgpu_type_mapping:
                raise exception.InvalidvGPUConfig(
                    reason="Duplicate types for PCI address %s"
                    % device_address
                )
            # Just checking whether the operator fat-fingered the address.
            # If it's wrong, it will return an exception
            try:
                # Validates whether it's a PCI ID...
                utils.parse_address(device_address)
            except exception.PciDeviceWrongAddressFormat:
                raise exception.InvalidvGPUConfig(
                    reason="Incorrect PCI address: %s" % device_address
                )
            pgpu_type_mapping[device_address] = vgpu_type
    return CONF.gpu_devices.enabled_vgpu_types, pgpu_type_mapping


def _get_vgpu_type_per_pgpu(device_address, supported_vgpu_types,
                            pgpu_type_mapping):
    """Provides the vGPU type the pGPU supports.

    :param device_address: the PCI device address in config,
                           eg.'0000:af:00.0'
    """
    supported_vgpu_types, pgpu_type_mapping = _get_supported_vgpu_types()
    # Bail out quickly if we don't support vGPUs
    if not supported_vgpu_types:
        LOG.warning('Unable to load vGPU_type from [gpu_devices] '
                    'Ensure "enabled_vgpu_types" is set if the gpu'
                    'is virtualized.')
        return

    try:
        # Validates whether it's a PCI ID...
        utils.parse_address(device_address)
    except (exception.PciDeviceWrongAddressFormat, IndexError):
        # this is not a valid PCI address
        LOG.warning("The PCI address %s was invalid for getting the"
                    "related vGPU type", device_address)
        return
    return pgpu_type_mapping.get(device_address)


def _discover_gpus(vendor_id):
    """param: vendor_id=VENDOR_ID means only discover Nvidia GPU on the host
    """
    # init vGPU conf
    cyborg.conf.devices.register_dynamic_opts(CONF)
    supported_vgpu_types, pgpu_type_mapping = _get_supported_vgpu_types()
    # discover gpu devices by "lspci"
    gpu_list = []
    gpus = gpu_utils.get_pci_devices(gpu_utils.GPU_FLAGS, vendor_id)
    # report trait,rc and generate driver object
    for gpu in gpus:
        m = gpu_utils.GPU_INFO_PATTERN.match(gpu)
        if m:
            gpu_dict = m.groupdict()
            # get hostname for deployable_name usage
            gpu_dict['hostname'] = CONF.host
            # get vgpu_type from cyborg.conf, otherwise vgpu_type=None
            vgpu_type = _get_vgpu_type_per_pgpu(
                gpu_dict["devices"], supported_vgpu_types, pgpu_type_mapping)
            # generate rc and trait for pGPU
            if not vgpu_type:
                gpu_dict["rc"] = constants.RESOURCES["PGPU"]
                traits = _get_traits(gpu_dict["vendor_id"],
                                     gpu_dict["product_id"])
            # generate rc and trait for vGPU
            else:
                # get rc
                gpu_dict["rc"] = constants.RESOURCES["VGPU"]
                mdev_path = os.path.expandvars(
                    '/sys/bus/pci/devices/{0}/mdev_supported_types'.
                    format(gpu_dict["devices"]))
                valid_types = os.listdir(mdev_path)
                if vgpu_type not in valid_types:
                    raise exception.InvalidVGPUType(name=vgpu_type)
                gpu_dict["vGPU_type"] = vgpu_type
                vGPU_path = os.path.expandvars(
                    '/sys/bus/pci/devices/{0}/mdev_supported_types/{1}/'
                    .format(gpu_dict["devices"], gpu_dict["vGPU_type"]))
                # transfer vgpu_type to vgpu_type_name.
                # eg. transfer 'nvidia-223' to 'T4_1B'
                with open(vGPU_path + 'name', 'r') as f:
                    name = f.read().strip()
                vgpu_type_name = name.split(' ')[1].replace('-', '_')
                traits = _get_traits(gpu_dict["vendor_id"],
                                     gpu_dict["product_id"],
                                     vgpu_type_name)
            gpu_dict.update(traits)
            gpu_list.append(_generate_driver_device(gpu_dict))
    return gpu_list


def discover(vendor_id):
    devs = _discover_gpus(vendor_id)
    return devs



# Copyright 2019 Intel, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import copy

from oslo_serialization import jsonutils

from cyborg.accelerator.drivers.fake import FakeDriver
from cyborg.tests import base


class TestFakeDriver(base.TestCase):

    def setUp(self):
        super(TestFakeDriver, self).setUp()

    def tearDown(self):
        super(TestFakeDriver, self).tearDown()

    def _get_pci_bdfs(self, num_pci_bdfs):
        """Return a list of PCI BDFs, first for cpid, rest
        for attach handles.
        """
        pci_bdf_template = {
            'domain': '0000', 'bus': '0c', 'device': '00', }
        pci_bdf_list = []
        for fn in range(num_pci_bdfs):
            pci_bdf = copy.deepcopy(pci_bdf_template)
            pci_bdf['function'] = str(fn)
            pci_bdf_list.append(jsonutils.dumps(pci_bdf))
        return pci_bdf_list

    def _get_cpid_attach_handles(self):
        num_accelerators = 2  # keep it simple
        pci_bdf_list = self._get_pci_bdfs(num_accelerators + 1)
        cpid = {
            'cpid_type': 'PCI',
            'cpid_info': pci_bdf_list[0],
        }

        ah_template = {
            'attach_type': 'TEST_PCI',
            'in_use': False,
        }
        ah1 = copy.deepcopy(ah_template)
        ah1['attach_info'] = pci_bdf_list[1]
        ah2 = copy.deepcopy(ah_template)
        ah2['attach_info'] = pci_bdf_list[2]
        return cpid, [ah1, ah2]

    def _validate_attach_handle_list(self, ah_list):
        self.assertTrue(all(
            ah['attach_type'] == 'TEST_PCI' for ah in ah_list))
        self.assertTrue(all(ah['in_use'] is False for ah in ah_list))

        ah_bdfs = [jsonutils.loads(ah['attach_info']) for ah in ah_list]
        bdf_list = [(bdf['domain'], bdf['bus'],
                     bdf['device'], bdf['function']) for bdf in ah_bdfs]
        domains = [bdf[0] for bdf in bdf_list]
        buses = [bdf[1] for bdf in bdf_list]
        pci_devs = [bdf[2] for bdf in bdf_list]
        pci_fns = [bdf[3] for bdf in bdf_list]

        # Ensure BDFs are unique
        self.assertEqual(len(set(bdf_list)), len(bdf_list),
                         "BDFs must be unique.")
        self.assertEqual(
            len(set(domains)), 1,
            "PCI domains are expected to be same for #accelerators < 256.")
        self.assertEqual(
            len(set(buses)), 1,
            "PCI bus #s are expected to be same for #accelerators < 256.")

        self.assertTrue(
            all([0 <= int(pci_dev) < 32 for pci_dev in pci_devs]),
            "PCI device numbers must be between 0 and 31."
            )
        self.assertTrue(
            all([0 <= int(pci_fn) < 8 for pci_fn in pci_fns]),
            "PCI function numbers must be between 0 and 7."
            )

    def test_discover(self):
        cpid, ah_list = self._get_cpid_attach_handles()
        # This is an example device from the fake driver.
        # We set num_accelerators = 2 for simplicity.
        expected = {'vendor': '0xABCD',
                    'type': 'FPGA',
                    'model': 'miss model info',
                    'deployable_list':
                        [{'num_accelerators': 2,
                          'name': 'FakeDevice',
                          'attach_handle_list': ah_list,
                          },
                         ],
                    'controlpath_id': cpid,
                    }

        fake_drv = FakeDriver()
        devices = fake_drv.discover()
        self.assertEqual(1, len(devices))
        for field in ['vendor', 'type', 'model']:
            self.assertEqual(devices[0][field], expected[field])

        expected_cpid = expected['controlpath_id']
        cpid = devices[0]['controlpath_id']
        self.assertEqual(cpid['cpid_type'], expected_cpid['cpid_type'])
        self.assertEqual(
            jsonutils.loads(cpid['cpid_info']),
            jsonutils.loads(expected_cpid['cpid_info']))

        deployables = devices[0].deployable_list
        self.assertEqual(1, len(deployables))
        self.assertEqual('fake-mini_FakeDevice', deployables[0]['name'])
        self.assertGreater(deployables[0]['num_accelerators'], 1)

        # Since num_accelerators can change, we don't test for its value.
        ah_list = deployables[0]['attach_handle_list']
        self.assertEqual(deployables[0]['num_accelerators'], len(ah_list))
        self._validate_attach_handle_list(ah_list)

    def test_generate_attach_handles(self):
        fake_drv = FakeDriver()
        pci_addr = '{"domain":"0000","bus":"0c","device":"00","function":"0"}'
        pci = {'slot': pci_addr}
        num_accelerators = 255  # Largest value that will fit within a bus
        ah_list = fake_drv._generate_attach_handles(pci, num_accelerators)

        self.assertEqual(num_accelerators, len(ah_list))
        self._validate_attach_handle_list(ah_list)



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Client side of the conductor RPC API."""

from oslo_config import cfg
import oslo_messaging as messaging

from cyborg.common import constants
from cyborg.common import rpc
from cyborg.objects import base as objects_base


CONF = cfg.CONF


class ConductorAPI(object):
    """Client side of the conductor RPC API.

    API version history:

    |    1.0 - Initial version.

    """

    RPC_API_VERSION = '1.0'

    def __init__(self, topic=None):
        super(ConductorAPI, self).__init__()
        self.topic = topic or constants.CONDUCTOR_TOPIC
        target = messaging.Target(topic=self.topic,
                                  version='1.0')
        serializer = objects_base.CyborgObjectSerializer()
        self.client = rpc.get_client(target,
                                     version_cap=self.RPC_API_VERSION,
                                     serializer=serializer)

    def report_data(self, context, hostname, driver_device_list):
        """Signal to conductor service to update the cyborg DB
        :parma context: request context.
        """
        cctxt = self.client.prepare(topic=self.topic)
        cctxt.call(context, 'report_data', hostname=hostname,
                   driver_device_list=driver_device_list)

    def device_profile_create(self, context, obj_devprof):
        """Signal to conductor service to create a device_profile.

        :param context: request context.
        :param obj_devprof: a created (but not saved) device_profile object.
        :returns: created device_profile object.
        """
        cctxt = self.client.prepare(topic=self.topic)
        return cctxt.call(context, 'device_profile_create',
                          obj_devprof=obj_devprof)

    def device_profile_delete(self, context, obj_devprof):
        """Signal to conductor service to delete a device_profile.
        :param context: request context.
        :param obj_devprof: a device_profile object to delete.
        """
        cctxt = self.client.prepare(topic=self.topic)
        cctxt.call(context, 'device_profile_delete',
                   obj_devprof=obj_devprof)

    def arq_create(self, context, obj_extarq, devprof_id):
        """Signal to conductor service to create an accelerator requests.

        :param context: request context.
        :param obj_extarq: a created (but not saved) accelerator_requests
        object
        :param devprof_id: a device profile id
        :returns: saved accelerator_requests object.
        """
        cctxt = self.client.prepare(topic=self.topic)
        return cctxt.call(context, 'arq_create', obj_extarq=obj_extarq,
                          devprof_id=devprof_id)

    def arq_delete_by_uuid(self, context, arqs):
        """Signal to conductor service to delete accelerator requests by
        ARQ UUIDs.

        :param context: request context.
        :param arqs: ARQ UUIDs joined with ','
        """
        cctxt = self.client.prepare(topic=self.topic)
        cctxt.call(context, 'arq_delete_by_uuid', arqs=arqs)

    def arq_delete_by_instance_uuid(self, context, instance):
        """Signal to conductor service to delete accelerator requests by
        instance UUID.

        :param context: request context.
        :param instance: UUID of instance whose ARQs need to be deleted
        """
        cctxt = self.client.prepare(topic=self.topic)
        cctxt.call(context, 'arq_delete_by_instance_uuid', instance=instance)

    def arq_apply_patch(self, context, patch_list, valid_fields):
        """Signal to conductor service to apply patch accelerator requests.

        :param context: request context.
        :param patch_list: A map from ARQ UUIDs to their JSON patches
        :param valid_fields: Dict of valid fields
        """
        cctxt = self.client.prepare(topic=self.topic)
        return cctxt.call(context, 'arq_apply_patch', patch_list=patch_list,
                          valid_fields=valid_fields)



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from cyborg.accelerator.drivers.spdk.util import common_fun
from cyborg.accelerator.drivers.spdk.util.pyspdk.vhost_client import VhostTgt
from cyborg.accelerator.drivers.spdk.vhost.vhost import VHOSTDRIVER
from cyborg.tests import base


class TestVHOSTDRIVER(base.TestCase):

    def setUp(self):
        super(TestVHOSTDRIVER, self).setUp()
        self.vhost_driver = VHOSTDRIVER()

    def tearDown(self):
        super(TestVHOSTDRIVER, self).tearDown()
        self.vhost_driver = None

    @mock.patch.object(VHOSTDRIVER, 'get_one_accelerator')
    def test_discover_accelerator(self, mock_get_one_accelerator):
        expect_accelerator = {
            'server': 'vhost',
            'bdevs': [{"num_blocks": 131072,
                       "name": "nvme1",
                       "block_size": 512
                       }],
            'scsi_devices': [],
            'luns': [{"claimed": True,
                      "name": "Malloc0"}],
            'interfaces': [{"core": 0,
                            "nqn": "nqn.2018-01.org.nvmexpress.discovery",
                            "hosts": []
                            }]
        }
        alive = mock.Mock(return_value=True)
        self.vhost_driver.py.is_alive = alive
        check_error = mock.Mock(return_value=True)
        common_fun.check_for_setup_error = check_error
        self.assertFalse(
            mock_get_one_accelerator.called,
            "Failed to discover_accelerator if py not alive."
        )
        acce_client = VhostTgt(self.vhost_driver.py)
        bdevs_fake = [{"num_blocks": 131072,
                       "name": "nvme1",
                       "block_size": 512
                       }]
        bdev_list = mock.Mock(return_value=bdevs_fake)
        acce_client.get_bdevs = bdev_list
        scsi_devices_fake = []
        scsi_device_list = mock.Mock(return_value=scsi_devices_fake)
        acce_client.get_scsi_devices = scsi_device_list
        luns_fake = [{"claimed": True,
                      "name": "Malloc0"}]
        lun_list = mock.Mock(return_value=luns_fake)
        acce_client.get_luns = lun_list
        interfaces_fake = \
            [{"core": 0,
              "nqn": "nqn.2018-01.org.nvmexpress.discovery",
              "hosts": []
              }]
        interface_list = mock.Mock(return_value=interfaces_fake)
        acce_client.get_interfaces = interface_list
        accelerator_fake = {
            'server': self.vhost_driver.SERVER,
            'bdevs': acce_client.get_bdevs(),
            'scsi_devices': acce_client.get_scsi_devices(),
            'luns': acce_client.get_luns(),
            'interfaces': acce_client.get_interfaces()
        }
        success_send = mock.Mock(return_value=accelerator_fake)
        self.vhost_driver.get_one_accelerator = success_send
        accelerator = self.vhost_driver.discover_accelerator()
        self.assertEqual(accelerator, expect_accelerator)

    def test_accelerator_list(self):
        expect_accelerators = [{
            'server': 'vhost',
            'bdevs': [{"num_blocks": 131072,
                       "name": "nvme1",
                       "block_size": 512
                       }],
            'scsi_devices': [],
            'luns': [{"claimed": True,
                      "name": "Malloc0"}],
            'interfaces': [{"core": 0,
                            "nqn": "nqn.2018-01.org.nvmexpress.discovery",
                            "hosts": []
                            }]
        },
            {
                'server': 'vhost_tgt',
                'bdevs': [{"num_blocks": 131072,
                           "name": "nvme1",
                           "block_size": 512
                           }],
                'scsi_devices': [],
                'luns': [{"claimed": True,
                          "name": "Malloc0"}],
                'interfaces': [{"core": 0,
                                "nqn": "nqn.2018-01.org.nvmexpress.discovery",
                                "hosts": []
                                }]
            }
        ]
        success_send = mock.Mock(return_value=expect_accelerators)
        self.vhost_driver.get_all_accelerators = success_send
        self.assertEqual(self.vhost_driver.accelerator_list(),
                         expect_accelerators)

    def test_install_accelerator(self):
        pass

    def test_uninstall_accelerator(self):
        pass

    def test_update(self):
        pass

    def test_attach_instance(self):
        pass

    def test_detach_instance(self):
        pass

    def test_delete_ip_address(self):
        pass

    def test_add_ip_address(self):
        pass



# Copyright 2020 ZTE Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for manipulating DeviceProfile via the DB API"""

import json
import sys

from oslo_utils import uuidutils

from cyborg.common import exception
from cyborg.tests.unit.db import base
from cyborg.tests.unit.db import utils


class TestDbDeviceProfile(base.DbTestCase):

    def test_create_dp(self):
        created_dp = utils.create_test_device_profile(self.context)
        expected_dp = utils.get_test_device_profile()
        self.assertEqual(json.loads(created_dp.profile_json)['groups'],
                         json.loads(expected_dp['profile_json'])['groups'])

    def test_create_dp_with_duplicate_name(self):
        utils.create_test_device_profile(self.context)
        duplicate_dp = utils.get_test_device_profile()
        duplicate_dp['id'] = 2
        duplicate_dp['uuid'] = uuidutils.generate_uuid()
        self.assertRaises(exception.DuplicateDeviceProfileName,
                          self.dbapi.device_profile_create,
                          self.context, duplicate_dp)

    def test_create_dp_with_duplicate_uuid(self):
        utils.create_test_device_profile(self.context)
        duplicate_dp = utils.get_test_device_profile()
        self.assertRaises(exception.DeviceProfileAlreadyExists,
                          self.dbapi.device_profile_create,
                          self.context, duplicate_dp)

    def test_get_by_uuid(self):
        created_dp = utils.create_test_device_profile(self.context)
        queried_dp = self.dbapi.device_profile_get_by_uuid(
            self.context, created_dp['uuid'])
        self.assertEqual(created_dp['uuid'], queried_dp['uuid'])
        self.assertIn('description', queried_dp)

    def test_get_by_id(self):
        created_dp = utils.create_test_device_profile(self.context)
        queried_dp = self.dbapi.device_profile_get_by_id(
            self.context, created_dp['id'])
        self.assertEqual(created_dp['id'], queried_dp['id'])
        self.assertIn('description', queried_dp)

    def test_update_with_name(self):
        created_dp = utils.create_test_device_profile(self.context)
        queried_dp = self.dbapi.device_profile_update(
            self.context, created_dp['uuid'], {'name': 'updated_name'})
        self.assertEqual('updated_name', queried_dp['name'])

    def test_update_with_description(self):
        created_dp = utils.create_test_device_profile(self.context)
        queried_dp = self.dbapi.device_profile_update(
            self.context, created_dp['uuid'], {'description': 'fake-desc'})
        self.assertEqual('fake-desc', queried_dp['description'])

    def test_list(self):
        uuids = []
        for i in range(1, 4):
            dp = utils.create_test_device_profile(
                self.context,
                id=i,
                uuid=uuidutils.generate_uuid(),
                name="device_profile_name_%s" % i)
            uuids.append(dp['uuid'])
        dps = self.dbapi.device_profile_list(self.context)
        dp_uuids = [item.uuid for item in dps]
        self.assertEqual(sorted(uuids), sorted(dp_uuids))

    def test_list_filter_by_name(self):
        utils.create_test_device_profile(
            self.context,
            id=1,
            uuid=uuidutils.generate_uuid(),
            name="name_1")
        utils.create_test_device_profile(
            self.context,
            id=2,
            uuid=uuidutils.generate_uuid(),
            name="name_2")
        res = self.dbapi.device_profile_list_by_filters(
            self.context, filters={"name": "name_1"})
        self.assertEqual(1, len(res))
        self.assertEqual('name_1', res[0]['name'])

    def test_delete(self):
        created_dp = utils.create_test_device_profile(self.context)
        return_value = self.dbapi.device_profile_delete(
            self.context,
            created_dp['uuid'])
        self.assertIsNone(return_value)

    def test_get_by_uuid_not_exist(self):
        random_uuid = uuidutils.generate_uuid()
        self.assertRaises(exception.ResourceNotFound,
                          self.dbapi.device_profile_get_by_uuid,
                          self.context, random_uuid)

    def test_list_filter_is_none(self):
        """The main test is filters=None. If filters=None,
        it will be initialized to {}, that will return all device
        profiles same as the List Device Profiles API response.
        """
        utils.create_test_device_profile(
            self.context,
            id=1,
            uuid=uuidutils.generate_uuid(),
            name="foo_dp")
        res = self.dbapi.device_profile_list_by_filters(
            self.context, filters=None)
        self.assertEqual(1, len(res))
        self.assertEqual('foo_dp', res[0]['name'])

    def test_update_with_uuid_not_exist(self):
        utils.create_test_device_profile(self.context)
        random_uuid = uuidutils.generate_uuid()
        self.assertRaises(exception.ResourceNotFound,
                          self.dbapi.device_profile_update,
                          self.context,
                          random_uuid,
                          {'name': 'updated_name'})

    def test_get_by_id_not_exist(self):
        fake_id = sys.maxsize
        self.assertRaises(exception.ResourceNotFound,
                          self.dbapi.device_profile_get_by_id,
                          self.context, fake_id)

    def test_get_by_name_not_exist(self):
        random_name = 'fake' + uuidutils.generate_uuid()
        self.assertRaises(exception.ResourceNotFound,
                          self.dbapi.device_profile_get,
                          self.context, random_name)

    def test_delete_by_uuid_not_exist(self):
        random_uuid = uuidutils.generate_uuid()
        self.assertRaises(exception.ResourceNotFound,
                          self.dbapi.device_profile_delete,
                          self.context, random_uuid)
